import util.io.env

import util.logging
logger = util.logging.logger


BATCH_SYSTEM_ENV_NAME = 'BATCH_SYSTEM'
try:
    BATCH_SYSTEM_STR = util.io.env.load(BATCH_SYSTEM_ENV_NAME)
    IS_RZ = BATCH_SYSTEM_STR == 'RZ-PBS'
    IS_NEC = BATCH_SYSTEM_STR == 'NEC-NQSII'
    IS_NONE = False
except util.io.env.EnvironmentLookupError:
    IS_RZ = False
    IS_NEC = False
    IS_NONE = True

if IS_RZ:
    logger.debug('Choosing batch system {}.'.format(BATCH_SYSTEM_STR))
    from util.batch.rz.system import *
elif IS_NEC:
    logger.debug('Choosing batch system {}.'.format(BATCH_SYSTEM_STR))
    from util.batch.nec.system import *
elif IS_NONE:
    logger.warn('Environmental variable {} is not set. Chosing general batch system.'.format(BATCH_SYSTEM_ENV_NAME))
    from util.batch.general.system import *
else:
    raise ValueError('Batch system {} is unknown.'.format(BATCH_SYSTEM_STR))



import numpy as np
import scipy.sparse
import scipy.sparse.sputils

import util.math.util
import util.logging
logger = util.logging.logger



def csr_matrix(value_function, shape, value_range, dtype=None, number_of_processes=None, chunksize=None):
    import multiprocessing

    assert callable(value_function)

    ## prepare values for pool
    if number_of_processes is None:
        number_of_processes = multiprocessing.cpu_count()
    if chunksize is None:
        chunksize = max([int(len(value_range) / (number_of_processes*10**3)), 1])

    ## get data
    logger.debug('Creating crs matrix of shape {} and type {} with {} processes and chunksize {}.'.format(shape, dtype, number_of_processes, chunksize))

    ## add values to matrix
    def sum_values_to_csr_matrix(results):
        i = 0

        ## init matrix
        m = scipy.sparse.csr_matrix(shape, dtype=dtype)

        ## add results
        for m_i in results:
            logger.debug('Adding values for index {} to total matrix.'.format(i))
            m = m + m_i
            i = i+1

        return m

    ## parallel
    if number_of_processes > 1:
        with multiprocessing.pool.Pool(processes=number_of_processes) as pool:
            results = pool.imap(value_function, value_range, chunksize=chunksize)
            m = sum_values_to_csr_matrix(results)
    ## serial
    else:
        results = map(value_function, value_range)
        m = sum_values_to_csr_matrix(results)


    return m



def diag(d):
    n = len(d)
    D = scipy.sparse.dia_matrix((d[np.newaxis,:], [0]), shape=(n,n))
    return D



def list_to_array(values, dtype=None):
    # a = np.asarray(values, dtype=dtype) #-> numpy bug
    n = len(values)
    a = np.empty(n, dtype=dtype)
    for i in range(n):
        a[i] = values[i]
    return a




class InsertableMatrix():

    def __init__(self, shape, dtype=np.float64):
        self.shape = shape
        self.data_dtype = dtype
        self.row_indices = []
        self.colum_indices = []
        self.data = []
        self.nnz = 0
        logger.debug('Initiating insertable matrix with shape {} and data dtype {}.'.format(self.shape, self.data_dtype))


    def insert(self, i, j, v):
        self.row_indices.append(i)
        self.colum_indices.append(j)
        self.data.append(v)
        assert len(self.row_indices) == len(self.colum_indices) == len(self.data)
        assert len(self.data) == self.nnz + 1
        self.nnz = self.nnz + 1
    
    
    def __setitem__(self, index, value):
        i, j = index
        self.insert(i, j, value)


    def tocoo(self):
        logger.debug('Prepare coo matrix with {} entries and data dtype {}.'.format(len(self.data), self.data_dtype))
        assert len(self.row_indices) == len(self.colum_indices) == len(self.data)

        indices_dtype = util.math.util.min_int_dtype(*self.shape)
        row = list_to_array(self.row_indices, dtype=indices_dtype)
        col = list_to_array(self.colum_indices, dtype=indices_dtype)
        data = list_to_array(self.data, dtype=self.data_dtype)

        matrix = scipy.sparse.coo_matrix((data, (row, col)), shape=self.shape)
        if matrix.dtype != self.data_dtype:
            matrix = matrix.astype(self.data_dtype)

        logger.debug('Returning matrix {!r}.'.format(matrix))
        assert matrix.dtype == self.data_dtype
        return matrix


    def asformat(self, format='csc'):
        logger.debug('Prepare {} matrix with {} entries and dtype {}.'.format(format, len(self.data), self.data_dtype))
        matrix = self.coo_matrix().asformat(format)
        if matrix.dtype != self.data_dtype:
            matrix = matrix.astype(self.data_dtype)

        logger.debug('Returning matrix {!r}.'.format(matrix))
        assert matrix.dtype == self.data_dtype
        return matrix





import itertools

import cachetools
import cachetools.keys


_kwargs_mark = (object(),)

def hash_key(*args, **kwargs):
    """Return a cache key for the specified hashable arguments."""

    if kwargs:
        return cachetools.keys._HashedTuple(args + _kwargs_mark + tuple(itertools.chain(sorted(kwargs.items()))))
    else:
        return cachetools.keys._HashedTuple(args)


def typed_hash_key(*args, **kwargs):
    """Return a typed cache key for the specified hashable arguments."""

    key = hash_key(*args, **kwargs)
    key += tuple(type(v) for v in args)
    key += tuple(type(v) for _, v in sorted(kwargs.items()))
    return key



_dependency_mark = (object(),)

def attribute_dependend_key(key_function, *dependencies):
    """Return a cache key for the specified hashable arguments with additional dependent arguments."""

    def dependend_key_function(self, *args, **kwargs):
        key = hash_key(*args, **kwargs)
        
        if len(dependencies) > 0:
            dependecies_dict = {}
            for dependency in dependencies:
                value = eval(dependency)
                dependecies_dict[dependency] = value
            
            key = key + cachetools.keys._HashedTuple(_dependency_mark + tuple(itertools.chain(sorted(dependecies_dict.items()))))

        return key
    
    return dependend_key_function


def attribute_dependend_hash_key(*dependencies):
    return attribute_dependend_key(hash_key, *dependencies)


def attribute_dependend_typed_hash_key(*dependencies):
    return attribute_dependend_key(typed_hash_key, *dependencies)


def decorator(maxsize=1, key=hash_key, lock=None, dependency=None):
    cache = cachetools.LRUCache(maxsize=maxsize)
    if dependency is not None:
        if isinstance(dependency, str):
            dependency = (dependency,)
        key = attribute_dependend_key(key, *dependency)
    return cachetools.cached(cache, key=key, lock=lock)







import numpy as np

import util.math.check


def lex_sorted_indices(a, order=1):
    util.math.check.ndim(a, 2)

    n = a.shape[1]
    if order > 0:
        axes = range(n)
    else:
        axes = range(n-1, -1, -1)

    return lex_sorted_indices_by_axes(a, axes)


def lex_sorted_indices_by_axes(a, axes):
    util.math.check.ndim(a, 2)

    lex_list = []
    for i in range(len(axes)-1, -1, -1):
        lex_list.append(a[:, axes[i]])
    sorted_indices = np.lexsort(lex_list)

    return sorted_indices



import util.logging
logger = util.logging.logger


class Observable:
    
    def __init__(self, observers=None):
        self.observers = []
        if observers is None:
            observers = []
        for observer in observers:
            self.add_observer(observer)
    
    def add_observer(self, observer):
        logger.debug('{}: observer {} added.'.format(self.__class__.__name__, observer))
        self.observers.append(observer)
    
    def remove_observer(self, observer):
        logger.debug('{}: observer {} removed.'.format(self.__class__.__name__, observer))
        self.observers.remove(observer)
    
    
    def _notify_observers(self, notify_method):
        for observer in self.observers:
            notify_method(observer)
    
    def notify_changed(self, key, new_value, old_value):
        logger.debug('{}: notify changed: key {}, value {}, old value {}.'.format(self.__class__.__name__, key, value, old_value))
        self._notify_observers(lambda observer: observer.notify_changed(key, new_value, old_value))
    
    def notify_added(self, key, new_value):
        logger.debug('{}: notify add: key {}, value {}.'.format(self.__class__.__name__, key, value))
        self._notify_observers(lambda observer: observer.notify_added(key, new_value))
    
    def notify_removed(self, key, old_value):
        logger.debug('{}: notify del: key {}, old value {}.'.format(self.__class__.__name__, key, old_value))
        self._notify_observers(lambda observer: observer.notify_removed(key, old_value))


import abc
import os
import re

import numpy as np

import util.io.fs
import util.index_database.general

import util.logging
logger = util.logging.logger



class Database(util.index_database.general.Database):
    
    def __init__(self, value_dir, value_filenames, value_reliable_decimal_places=15, tolerance_options=None):
        ## call super constructor
        super().__init__(value_reliable_decimal_places=value_reliable_decimal_places, tolerance_options=tolerance_options)
        
        ## set value dir
        self.value_dir = value_dir
        
        ## save filenames in list
        if isinstance(value_filenames, str):
            value_filenames = [value_filenames]
        self.value_filenames = value_filenames
        assert len(self.value_filenames) > 0
        
    
    def __str__(self):
        return 'Index file system database {}'.format(self.value_dir)
    
    
    ## setter and getter for files and dirs
    
    @property
    def value_dir(self):
        return self._value_dir
    
    @value_dir.setter
    def value_dir(self, value_dir):               
        ## check value dir
        for s in ['{', '}']:
            if len(value_dir.split(s)) != 2:
                raise ValueError('The value dirname must contain exactly one "{}". But the filename is {}.'.format(s, value_file))
        
        ## calculate dirs
        value_dir = os.path.realpath(value_dir)
        
        base_dir_including_index = value_dir
        while re.search('\{.*\}', os.path.dirname(base_dir_including_index)):
            base_dir_including_index = os.path.dirname(base_dir_including_index)
        
        base_dir = os.path.dirname(base_dir_including_index)
        assert len(base_dir) > 0

        ## make base dir
        os.makedirs(base_dir, exist_ok=True)
        
        ## store needed dirs
        self._value_dir = value_dir
        self._base_dir_including_index = base_dir_including_index
        self._base_dir = base_dir

        ## save needed regular expressions        
        index_dir = os.path.basename(base_dir_including_index)
        self._index_dir_regular_expression_search = re.sub('\{.*\}', '[0-9]*', index_dir)
        self._index_dir_regular_expression_split = re.sub('\{.*\}', '', index_dir)
    
    
    @property
    def value_filenames(self):
        return self._value_filenames
    
    @value_filenames.setter
    def value_filenames(self, value_filenames):        
        ## save filenames in list
        if isinstance(value_filenames, str):
            value_filenames = [value_filenames]
        assert len(value_filenames) > 0
        
        self._value_filenames = value_filenames
    
    
    ## load and save

    @abc.abstractmethod
    def _load_file(self, value_file):
        raise NotImplementedError()

    @abc.abstractmethod
    def _save_file(self, value_file, value):
        raise NotImplementedError()

    
    ## access
    
    def value_files(self, index):
        ## make absolute filenames
        value_dir = self.value_dir.format(index)
        value_files = [os.path.join(value_dir, value_filename) for value_filename in self.value_filenames]
        return value_files
        
    
    def get_value(self, index):
        value_files = self.value_files(index)

        ## load each file
        value_array = []
        for value_file in value_files:        
            try:
                current_value_array = self._load_file(value_file)
            except FileNotFoundError:
                raise util.index_database.general.DatabaseIndexError(index)
            value_array.append(current_value_array)

        ## return as one array
        if len(value_array) > 1:
            value_array = np.asanyarray(value_array)
        else:
            value_array = value_array[0]
        return value_array


    def set_value(self, index, value, overwrite=False):
        logger.debug('{}: Setting value at index {} to {} with overwrite {}.'.format(self, index, value, overwrite))
        
        ## check value and make it as two dim array
        value = np.asanyarray(value)        
        assert (len(self.value_filenames) == 1 and (value.ndim == 1 or (value.ndim == 2 and len(value) == len(self.value_filenames)))) or (len(self.value_filenames) > 1 and len(value) == len(self.value_filenames))
        
        if value.ndim == 1:
            value = [value]

        ## make value dir
        value_dir = self.value_dir.format(index)
        os.makedirs(value_dir, exist_ok=True)

        ## save each value to file
        for i in range(len(self.value_filenames)):
            value_file = os.path.join(value_dir, self.value_filenames[i])
            value_file_exists = os.path.exists(value_file)
            if value_file_exists and overwrite:
                util.io.fs.make_writable(value_file)
            if overwrite or not value_file_exists:
                self._save_file(value_file, value[i])
                util.io.fs.make_read_only(value_file)
            else:
                logger.debug('{}: Overwritting value at index {} is not allowed.'.format(self, index))
                raise util.index_database.general.DatabaseOverwriteError(index)
    
    
    def used_indices(self):
        ## get all value files
        all_index_dirs = util.io.fs.find_with_regular_expression(self._base_dir, self._index_dir_regular_expression_search, exclude_files=True, use_absolute_filenames=False, recursive=False)
        
        ## get all used indices
        used_indices = []
        for index_dir in all_index_dirs:
            # get name starting at index
            extracted = re.split(self._index_dir_regular_expression_split, index_dir)
            assert len(extracted) in [2, 3]
            extracted = extracted[1]
            # get index only
            extracted = re.search(r'\d+', extracted).group()
            # append int index
            index = int(extracted)
            used_indices.append(index)
        
        logger.debug('{}: Got {} used indices.'.format(self, len(used_indices)))
        return used_indices
    
    
    def remove_index(self, index, force=False):
        logger.debug('{}: Removing index {}.'.format(self, index))
        
        ## get topmost dir with index
        base_dir_including_index = self._base_dir_including_index.format(index)

        ## remove directory
        logger.debug('{}: Removing value dir {}.'.format(self, base_dir_including_index))
        assert base_dir_including_index.startswith(self._base_dir) and len(base_dir_including_index) > len(self._base_dir)
        try:
            util.io.fs.remove_recursively(base_dir_including_index, force=force, exclude_dir=False)
        except FileNotFoundError:
            raise util.index_database.general.DatabaseIndexError(index)
            



import numpy as np

import util.math.sort
import util.io.object
import util.logging

logger = util.logging.logger


def _isdict(d):
    if isinstance(d, dict):
        return True
    else:
        from blist import sorteddict
        return isinstance(d, sorteddict)


class MultiDict():

    def __init__(self, sorted=False):
        if sorted:
            from blist import sorteddict
            self._value_dict = sorteddict()
        else:
            self._value_dict = dict()

    ## properties

    @property
    def value_dict(self):
        return self._value_dict

    @property
    def sorted(self):
        return not isinstance(self.value_dict, dict)

    @property
    def SUPPORTED_RETURN_TYPES(self):
        return ('array', 'self', 'self_type', 'self_type_unsorted', 'self_type_sorted', 'multi_dict_unsorted', 'multi_dict_sorted')


    ## str

    def __str__(self):
        name = self.__class__.__name__
        if self.sorted:
            name = name + '(Sorted)'
        else:
            name = name + '(Unsorted)'
        return name


    ## mapping methods

    def get_value_list(self, key):
        value = self.value_dict
        try:
            for i in range(len(key)):
                if not _isdict(value):
                    raise KeyError('Key {} not applicable. Under key {} is a value of type {}.'.format(key, key[:i], type(value)))
                try:
                    value = value[key[i]]
                except TypeError as e:
                    raise KeyError('Key {} not applicable. The {}th value of the key is of type {}.'.format(key, i, type(key[i]))) from e
        except KeyError as e:
            # raise KeyError(key) from e
            value = []
        return value

    def __getitem__(self, key):
        return self.get_value_list(key)


    def _get_last_dict(self, key):
        value_dict = self.value_dict
        value_dict_type = type(value_dict)

        for i in range(len(key)-1):
            value_dict = value_dict.setdefault(key[i], value_dict_type())

        return value_dict


    def set_value_list(self, key, value_list):
        value_list = list(value_list)
        last_dict = self._get_last_dict(key)
        last_dict[key[len(key)-1]] = value_list

    def __setitem__(self, key, value):
        self.set_value_list(key, value)


    def has_values(self, key):
        try:
            value_list = self.get_value_list(key)
        except KeyError:
            return False
        return len(value_list) > 0

    def __contains__(self, key):
        return self.has_values(key)


    @property
    def len(self):
        total_len = 0
        for (key, value_list) in self.iterator_keys_and_value_lists():
            total_len += len(value_list)

        return total_len

    def __len__(self):
        return self.len



    ## add

    def _get_or_init_value_list(self, key):
        last_dict = self._get_last_dict(key)
        value_list = last_dict.setdefault(key[len(key)-1], [])
        return value_list

    def extend_value_list(self, key, value_list):
        if len(value_list) > 0:
            self._get_or_init_value_list(key).extend(value_list)

    def append_value(self, key, value):
        self._get_or_init_value_list(key).append(value)

    def _add_value_lists(self, keys, value_lists, add_function):
        assert callable(add_function)
        if keys is None:
            keys = []
        if value_lists is None:
            value_lists = []

        if len(keys) != len(value_lists):
            raise ValueError('Len of keys {} and len of values {} have to be the same!'.format(len(keys), len(value_lists)))

        values_len = len(value_lists)
        logger.debug('Adding {} values.'.format(values_len))
        for i in range(values_len):
            add_function(keys[i], value_lists[i])

    def extend_value_lists(self, keys, value_lists):
        add_function = lambda key, value_list: self.extend_value_list(key, value_list)
        self._add_value_lists(keys, value_lists, add_function)

    def append_values(self, keys, values):
        add_function = lambda key, value: self.append_value(key, value)
        self._add_value_lists(keys, values, add_function)


    ## remove
    
    def clear(self):
        self._value_dict = type(self.value_dict)()

    def remove_value(self, key, value):
        logger.debug('Removing value {} for key {}.'.format(value, key))
        value_list = self.get_value_list(key)
        n = len(value_list)
        value_list[:] = [v for v in value_list if not np.all(np.isclose(v, value))]
        if len(value_list) == n:
            raise KeyError('Value {} was not deposited for key {}.'.format(value, key))
        #TODO remove dict entries if list is empty
    
    def remove_values(self, multi_dict):
        logger.debug('Removing {} values.'.format(len(multi_dict)))
        for key, value_list in multi_dict.iterator_keys_and_value_lists():
            for value in value_list:
                self.remove_value(key, value)


    ## access

    def keys(self):
        all_keys = []
        for (key, values) in self.iterator_keys_and_value_lists():
            for value in values:
                all_keys.append(key)

        all_keys = np.array(all_keys)
        # assert all_keys.ndim == 2
        return all_keys


    def values(self):
        all_values = []
        for (key, values) in self.iterator_keys_and_value_lists():
            all_values.extend(values)

        all_values = np.array(all_values)
        return all_values


    def items(self):
        all_keys = self.keys()
        all_keys = all_keys.reshape((all_keys.shape[0], -1))
        all_values = self.values()
        if all_values.ndim == 1:
            all_values = all_values[:, np.newaxis]
        return np.concatenate([all_keys, all_values], axis=1)

    def toarray(self):
        return self.items()


    ## io
    def save(self, file, only_dict=True):
        logger.debug('Saving {} to {}.'.format(self, file))
        if only_dict:
            util.io.object.save(file, self.value_dict)
        else:
            util.io.object.save(file, self)


    # def load(self, file):
    #     logger.debug('Loading {} from {}.'.format(self, file))
    #     self._value_dict = util.io.io.load_object(file)
    #     return self

    @classmethod
    def load(cls, file):
        logger.debug('Loading {} from {}.'.format(cls.__name__, file))

        ## load object
        loaded_object = util.io.object.load(file)

        ## check if dict
        is_dict = isinstance(loaded_object, dict)
        if not is_dict:
            from blist import sorteddict
            is_dict = isinstance(loaded_object, sorteddict)

        ## if dict make new object with dict
        if is_dict:
            obj = cls()
            obj._value_dict = loaded_object
        ## otherwise return loaded object
        else:
            obj = loaded_object

        return obj


    ## properties

    @property
    def value_dict(self):
        return self._value_dict

    @property
    def sorted(self):
        return not isinstance(self.value_dict, dict)


        obj = cls()
        obj._value_dict = util.io.object.load(file)
        return obj



    ## create

    def new_like(self, sorted=None):
        if sorted is None:
            sorted = self.sorted
        new = type(self)(sorted=sorted)
        return new
    
    
    def copy(self):
        import copy
        return copy.deepcopy(self)


    def _return_items_as_type(self, keys, value_lists, return_type=None):
        logger.debug('Returning {} values as type {}.'.format(len(keys), return_type))

        ## chose default
        if return_type is None:
            return_type = 'array'
        if return_type == 'multi_dict':
            return_type = 'multi_dict_unsorted'


        ## check input
        if return_type not in self.SUPPORTED_RETURN_TYPES:
            raise ValueError('Unknown return_type "{}". Only {} are supported.'.format(return_type, self.SUPPORTED_RETURN_TYPES))

        n = len(keys)
        if n != len(value_lists):
            raise ValueError('Len of keys {} and len of value lists {} have to be the same!'.format(len(keys), len(value_lists)))


        ## make list of value lists
        value_lists = list(value_lists)
        for i in range(n):
            try:
                value_lists[i] = list(value_lists[i])
            except TypeError:
                value_lists[i] = [value_lists[i]]


        ## return multi_dict type
        if return_type in ('self', 'self_type', 'self_type_unsorted', 'self_type_sorted', 'multi_dict_unsorted', 'multi_dict_sorted'):
            if return_type == 'self':
                m = self
                m.clear()
            if return_type in ('self_type', 'self_type_unsorted', 'self_type_sorted'):
                if return_type == 'self_type':
                    sorted = None
                if return_type == 'self_type_unsorted':
                    sorted = False
                if return_type == 'self_type_sorted':
                    sorted = True
                m = self.new_like(sorted=sorted)
            if return_type == 'multi_dict_unsorted':
                m = MultiDict(sorted=False)
            if return_type == 'multi_dict_sorted':
                m = MultiDict(sorted=True)

            m.extend_value_lists(keys, value_lists)
#             try:
#                 m.extend_value_lists(keys, value_lists)
#             except TypeError:
#                 m.append_values(keys, value_lists)
# #             m.append_values(keys, values)

            obj = m

        ## return array
        if return_type == 'array':
            if len(value_lists) > 0:
                def get_value_len(value):
                    try:
                        return len(value)
                    except TypeError:
                        return 1
                value_ref_len = get_value_len(value_lists[0][0])
                for value_list in value_lists:
                    for value in value_list:
                        if get_value_len(value) != value_ref_len:
                            raise ValueError('Len of each value has to be the same, but a len is {}!'.format(get_value_len(value)))

                n = len(keys)
                m = len(keys[0]) + value_ref_len

                array = np.empty((n, m))
                for i in range(n):
                    array[i, :-value_ref_len] = keys[i]
                    for value in value_lists[i]:
                        array[i, -value_ref_len:] = value

                obj = array
            else:
                obj = np.empty((0, 0))

        return obj


    ## iterate

    def _iterate_generator_value_dict(self, value_dict, value_dict_type=None, key_prefix=()):
        if value_dict_type is None:
            value_dict_type = type(value_dict)
        for (key, value) in value_dict.items():
            total_key = key_prefix + (key,)
            if isinstance(value, value_dict_type):
                yield from self._iterate_generator_value_dict(value, value_dict_type=value_dict_type, key_prefix=total_key)
            else:
                yield (total_key, value)


    def iterator_keys_and_value_lists(self):
        value_dict = self.value_dict
        yield from self._iterate_generator_value_dict(value_dict)


    def iterate_items(self, fun, min_number_of_values=1, return_type='array'):
        assert callable(fun)

        ## init
        new_keys = []
        new_values = []

        ## iterate
        for (key, values) in self.iterator_keys_and_value_lists():
            if len(values) >= min_number_of_values:
                key = np.asarray(key)
                values = np.asarray(values)
                new_value = fun(key, values)

                ## insert
                new_keys.append(key)
                new_values.append(new_value)

        ## finishing
        return self._return_items_as_type(new_keys, new_values, return_type=return_type)


    def iterate_values(self, fun, min_number_of_values=1, return_type='array'):
        fun_wrapper = lambda key, values: fun(values)
        return self.iterate_items(fun_wrapper, min_number_of_values=min_number_of_values, return_type=return_type)



    ## transform keys

    def transform_keys(self, transform_function):
        logger.debug('Transforming keys of {}.'.format(self))
        assert callable(transform_function)

        value_dict = self._value_dict
        self.clear()

        for (key, value_list) in self._iterate_generator_value_dict(value_dict):
            key_transformed = transform_function(key)
            self.extend_value_list(key_transformed, value_list)


    def keys_to_int_keys(self, dtype=np.int):
        logger.debug('Converting keys to type {}.'.format(dtype))
        transform_function = lambda key: tuple(np.array(np.round(key), dtype=dtype))
        self.transform_keys(transform_function)


    def dicard_key_dim(self, key_dim):
        logger.debug('Discarding key dim {}.'.format(key_dim))

        def transform_function(current_key):
            current_key = list(current_key)
            current_key[key_dim] = 0
            current_key = tuple(current_key)
            return current_key

        self.transform_keys(transform_function)


    def dicard_key_dims(self, key_dims):
        for key_dim in key_dims:
            self.dicard_key_dim(key_dim)



    ## transform values

    def transform_value_lists(self, transform_function):
        logger.debug('Transforming value lists of {}.'.format(self))
        assert callable(transform_function)

        for (key, value_list) in self.iterator_keys_and_value_lists():
            transformed_value_list = transform_function(key, value_list)
            self.set_value_list(key, transformed_value_list)

    def transform_values(self, transform_function):
        logger.debug('Transforming values of {}.'.format(self))
        assert callable(transform_function)

        def transform_function_wapper(key, value_list):
            transformed_value_list = []
            for value in value_list:
                transformed_value_list.append(transform_function(key, value))
            return transformed_value_list

        self.transform_value_lists(transform_function_wapper)

        # for (key, value_list) in self.iterator_keys_and_value_lists():
        #     transformed_value_list = []
        #     for value in value_list:
        #         transformed_value_list.append(transform_function(key, value))
        #     self.set_value_list(key, transformed_value_list)


    def set_min_value(self, min_value):
        logger.debug('Applying min value {} to values.'.format(min_value))

        transform_function = lambda key, value: max([value, min_value])
        self.transform_value(transform_function)


#     def is_at_least_value(self, value_threshold):
#         logger.debug('Applying is at least value {} to values.'.format(value_threshold))
#
#         transform_function = lambda key, value: value >= value_threshold)
#         self.transform_value(transform_function)


    def log_values(self):
        logger.debug('Applying logarithm to values.')

        transform_function = lambda key, value: np.log(value)
        self.transform_value(transform_function)



    ## filter

    def filter_with_boolean_function(self, boolean_filter_function, return_type='self'):
        assert callable(boolean_filter_function)

        filtered_keys = []
        filtered_value_lists = []

        for (key, value_list) in self.iterator_keys_and_value_lists():
            if boolean_filter_function(key, value_list):
                filtered_keys.append(key)
                filtered_value_lists.append(value_list)

        logger.debug('Filtered {} value lists.'.format(len(filtered_value_lists)))
        return self._return_items_as_type(filtered_keys, filtered_value_lists, return_type=return_type)


    def filter_min_number_of_values(self, min_number_of_values=1, return_type='self'):
        def boolean_filter_function(key, value_list):
            return len(value_list) >= min_number_of_values

        return self.filter_with_boolean_function(boolean_filter_function, return_type=return_type)


    def filter_key_range(self, key_index, bounds, return_type='self'):
        def boolean_filter_function(key, value_list):
            return bounds[0] <= key[key_index] <= bounds[1]

        return self.filter_with_boolean_function(boolean_filter_function, return_type=return_type)



    def filter_with_list_function(self, list_filter_function, return_type='self'):
        assert callable(list_filter_function)

        filtered_keys = []
        filtered_value_lists = []

        for (key, value_list) in self.iterator_keys_and_value_lists():
            filter_value_list = list_filter_function(key, value_list)
            if len(filter_value_list) > 0:
                filtered_keys.append(key)
                filtered_value_lists.append(filter_value_list)

        logger.debug('Filtered {} value lists.'.format(len(filtered_value_lists)))
        return self._return_items_as_type(filtered_keys, filtered_value_lists, return_type=return_type)


    def filter_finite(self, return_type='self'):
        def list_filter_function(key, value_list):
            filtered_value_list = []
            for value in value_list:
                if np.all(np.isfinite(value)):
                    filtered_value_list.append(value)
            return filtered_value_list

        return self.filter_with_list_function(list_filter_function, return_type=return_type)





    ## compute values

    def numbers(self, min_number_of_values=1, return_type='array'):
        logger.debug('Calculate numbers of values with at least {} values.'.format(min_number_of_values))

        return self.iterate_values(len, min_number_of_values, return_type=return_type)


    def means(self, min_number_of_values=1, min_value=0, return_type='array'):
        logger.debug('Calculate means of values with at least {} values with minimal mean {}.'.format(min_number_of_values, min_value))
        if min_value is None:
            min_value = - np.inf

        def calculate_function(values):
            mean = np.average(values)
            mean = max([mean, min_value])
            return mean

        return self.iterate_values(calculate_function, min_number_of_values, return_type=return_type)


    def variances(self, min_number_of_values=3, min_value=0, return_type='array'):
        logger.debug('Calculate variances of values with at least {} values with minimal variance {}.'.format(min_number_of_values, min_value))
        if min_value is None:
            min_value = 0

        def calculate_function(values):
            mean = np.average(values)
            number_of_values = values.size
            variance = np.sum((values - mean)**2) / (number_of_values - 1)
            variance = max([variance, min_value])
            return variance

        return self.iterate_values(calculate_function, min_number_of_values, return_type=return_type)


    def standard_deviations(self, min_number_of_values=3, min_value=0, return_type='array'):
        logger.debug('Calculate standard deviations of values with at least {} values with minimal deviation {}.'.format(min_number_of_values, min_value))
        if min_value is None:
            min_value = 0

        def calculate_function(values):
            mean = np.average(values)
            number_of_values = values.size
            deviation = (np.sum((values - mean)**2) / (number_of_values - 1))**(1/2)
            deviation = max([deviation, min_value])
            return deviation

        return self.iterate_values(calculate_function, min_number_of_values, return_type=return_type)




    ## tests for normality

    def dagostino_pearson_test(self, min_number_of_values=50, alpha=0.05, return_type='array'):
        logger.debug('Calculate D´Agostino-Person-test for normality of values with minimal {} values with alpha {}.'.format(min_number_of_values, alpha))
        import scipy.stats

        test_values = self.iterate_values(lambda x: scipy.stats.normaltest(x)[1], min_number_of_values, return_type=return_type)

        if alpha is not None:
            if return_type == 'array':
                test_values[:,-1] = (test_values[:,-1] >= alpha).astype(np.float)
            else:
                transform_function = lambda key, value: (value >= alpha).astype(np.float)
                self.transform_value(transform_function)

        return test_values


    def shapiro_wilk_test(self, min_number_of_values=50, alpha=0.05, return_type='array'):
        logger.debug('Calculate Shapiro-Wilk-test for normality of values with minimal {} values with alpha {}.'.format(min_number_of_values, alpha))
        import scipy.stats

        test_values = self.iterate_values(lambda x: scipy.stats.shapiro(x)[1], min_number_of_values, return_type=return_type)

        if alpha is not None:
            if return_type == 'array':
                test_values[:,-1] = (test_values[:,-1] >= alpha).astype(np.float)
            else:
                transform_function = lambda key, value: (value >= alpha).astype(np.float)
                self.transform_value(transform_function)

        return test_values


    def anderson_darling_test(self, min_number_of_values=50, alpha=0.05, return_type='array'):
        logger.debug('Calculate Anderson-Darling-test for normality of values with minimal {} values with alpha {}.'.format(min_number_of_values, alpha))
        import scipy.stats

        def test(x, alpha):
            ## get test values
            t = scipy.stats.anderson(x)
            test_value = t[0]
            test_bounds = t[1]
            test_alphas = t[2] / 100

            ## get bound for alpha
            index = np.where(test_alphas == alpha)[0]
            if len(index) == 0:
                raise ValueError('The alpha value {} is not supported for this test. Only to values {} are supported.'.format(alpha, test_alphas))
            index = index[0]
            bound = test_bounds[index]

            ## check if test passed
            return test_value <= bound

        test_values = self.iterate_values(lambda x:test(x, alpha), min_number_of_values, return_type=return_type)
        return test_values



class MultiDictPreprocessKey(MultiDict):

    def __init__(self, sorted=False):
        super().__init__(sorted=sorted)


    @staticmethod
    def _check_keys(keys):
        pass

    @staticmethod
    def _preprocess_keys(keys):
        raise NotImplementedError("Please implement this method.")


    ## override for key pairs

    def get_value_list(self, key):
        self._check_keys(key)
        key = self._preprocess_keys(key)
        return super().get_value_list(key)


    def set_value_list(self, key, value_list):
        self._check_keys(key)
        key = self._preprocess_keys(key)
        super().set_value_list(key, value_list)


    def _get_or_init_value_list(self, key):
        self._check_keys(key)
        key = self._preprocess_keys(key)
        return super()._get_or_init_value_list(key)




class MultiDictPointPairs(MultiDictPreprocessKey):

    def __init__(self, sorted=False):
        super().__init__(sorted=sorted)


    @staticmethod
    def _check_keys(keys):
        try:
            are_two_keys = len(keys) == 2
        except TypeError:
            raise ValueError('keys {} have to be a pair of keys.'.format(keys))

        try:
            are_two_keys = len(keys[0]) == len(keys[1])
        except (TypeError, KeyError):
            raise ValueError('keys {} have to be two keys of equal length.'.format(keys))


    def _iterate_generator_value_dict(self, value_dict, value_dict_type=None, key_prefix=()):
        if value_dict_type is None:
            value_dict_type = type(value_dict)
        for (key, value) in value_dict.items():
            total_key = key_prefix + (key,)
            if isinstance(value, value_dict_type):
                yield from self._iterate_generator_value_dict(value, value_dict_type, key_prefix=total_key)
            else:
                split_index = int(len(total_key)/2)
                yield ((total_key[0:split_index], total_key[split_index:]), value)


    def _return_items_as_type(self, keys, value_lists, return_type=None):
        if return_type is None or return_type == 'array':
            flattened_keys = []
            for key in keys:
                flattened_key = tuple(np.array(key).flatten())
                flattened_keys.append(flattened_key)
            keys = flattened_keys

        return super()._return_items_as_type(keys, value_lists, return_type=return_type)





class MultiDictPermutablePointPairs(MultiDictPointPairs):

    def __init__(self, sorted=False):
        super().__init__(sorted=sorted)
    #     self._pair_order = 1
    #
    #
    # @property
    # def pair_order(self):
    #     return self._pair_order

    # @staticmethod
    # def _check_keys(keys):
    #     try:
    #         are_two_keys = len(keys) == 2
    #     except TypeError:
    #         raise ValueError('keys {} have to be a pair of keys.'.format(keys))
    #
    #     try:
    #         are_two_keys = len(keys[0]) == len(keys[1])
    #     except (TypeError, KeyError):
    #         raise ValueError('keys {} have to be two keys of equal length.'.format(keys))


    @staticmethod
    def _preprocess_keys(keys):
        ## if two keys, make them to one sorted key
        key_array = np.array(keys)
        sorted_indices = util.math.sort.lex_sorted_indices(key_array, order=1)
        key_array = key_array[sorted_indices]
        keys = tuple(key_array.flat)
        return keys


    # def _iterate_generator_value_dict(self, value_dict, value_dict_type=None, key_prefix=()):
    #     if value_dict_type is None:
    #         value_dict_type = type(value_dict)
    #     for (key, value) in value_dict.items():
    #         total_key = key_prefix + (key,)
    #         if isinstance(value, value_dict_type):
    #             yield from self._iterate_generator_value_dict(value, value_dict_type, key_prefix=total_key)
    #         else:
    #             split_index = int(len(total_key)/2)
    #             yield ((total_key[0:split_index], total_key[split_index:]), value)
    #
    #
    # def _return_items_as_type(self, keys, value_lists, return_type=None):
    #     if return_type is None or return_type == 'array':
    #         flattened_keys = []
    #         for key in keys:
    #             flattened_key = tuple(np.array(key).flatten())
    #             flattened_keys.append(flattened_key)
    #         keys = flattened_keys
    #
    #     return super()._return_items_as_type(keys, value_lists, return_type=return_type)





class MultiDictDiffPointPairs(MultiDictPreprocessKey):

    def __init__(self, sorted=False):
        super().__init__(sorted=sorted)


    @staticmethod
    def _preprocess_keys(keys):
        ## if two keys, use diff as key
        key_array = np.array(keys)
        if key_array.ndim > 1:
            key = tuple(np.abs(key_array[1] - key_array[0]))
        else:
            key = tuple(np.abs(key_array))
        return key





import util.io.fs
import util.io.universal

import util.logging
logger = util.logging.logger



## file cache

def decorator(cache_file_function=None):
    
    def decorate_function(function, cache_file_function=None):
        ## if no cache file function is passed used passed cache file function name or default cache file function name 
        cache_file_function_defined = not (cache_file_function is None or isinstance(cache_file_function, str))
        if not cache_file_function_defined:
            ## passed cache file function name
            if isinstance(cache_file_function, str):
                cache_file_function_name = cache_file_function
            ## default cache file function name
            else:
                function_name = function.__name__
                cache_file_function_name = '{function_name}_cache_file'.format(function_name=function_name)
        
        def function_wrapper(*args, **kargs):            
            ## calculate cache file
            if cache_file_function_defined:
                cache_file = cache_file_function(*args, **kargs)
            else:
                cache_file = None
                try:
                    self = args[0]
                except IndexError:
                    logger.warn('This is not a method call! Cache file {} is not defined. Using no cache!'.format(cache_file_function_name))
                else:
                    try:
                        cache_file_function_by_name = getattr(self, cache_file_function_name)
                    except AttributeError:
                        logger.warn('Cache file {} is not defined in {}. Using no cache!'.format(cache_file_function_name, self))
                    else:
                        cache_file = cache_file_function_by_name(*args[1:], **kargs)
            
            ## if cache file is defined, use cache
            if cache_file is not None:                
                ## if cache file is available, load cached value
                try:
                    value = util.io.universal.load(cache_file)
                    logger.debug('Calculated value loaded from cache file {}.'.format(cache_file))
                ## if cache file is not available, calculate value and save as cached value
                except OSError:
                    logger.debug('No cached value found for {}.'.format(cache_file))
                    value = function(*args, **kargs)
                    logger.debug('Saving calculated value to cache file {}.'.format(cache_file))
                    util.io.universal.save(cache_file, value)
                    util.io.fs.make_read_only(cache_file)
            ## if cache file not defined, calculate value without cache
            else:
                value = function(*args, **kargs)
            
            return value
        
        return function_wrapper
    
    return lambda function: decorate_function(function, cache_file_function=cache_file_function)




import numpy as np



def calculate(f, x, f_x=None, typical_x=None, bounds=None, accuracy_order=2, eps=None, use_always_typical_x=True):
    x = np.asanyarray(x)
    
    ## init unpassed values
    if typical_x is None:
        typical_x = np.ones_like(x)
    elif not len(x) == len(typical_x):
        raise ValueError('x and typical_x must have the same length but their length are {} and {}.'.format(len(x), len(typical_x)))
    if bounds is None:
        bounds = ((-np.inf, np.inf),) * len(x)
    else:
        bounds = np.asanyarray(bounds)

    ## set h factors according to accuracy
    if accuracy_order == 1:
        h_factors = (1,)
        if eps is None:
            eps = np.spacing(1)**(1/2)
    elif accuracy_order == 2:
        h_factors = (1, -1)
        if eps is None:
            eps = np.spacing(1)**(1/3)
    else:
        raise ValueError('Accuracy order {} not supported.'.format(accuracy_order))
    
    ## calculate f(x) if needed
    if f_x is None and accuracy_order == 1:
        f_x =  f(x)

    ## init values
    n = len(x)
    m = len(h_factors)
    df = None

    ## for each x dim
    for i in range(n):
        h = np.empty(m, dtype=np.float64)
        
        ## for each h factor
        for j in range(m):
            ## calculate h
            if use_always_typical_x:
                typical_x_i = np.abs(typical_x[i])
            else:
                typical_x_i = np.max([np.abs(typical_x[i]), np.abs(x[i])])
            h[j] = h_factors[j] * eps * typical_x_i
            if accuracy_order == 1 and x[i] < 0:
                h[j] = - h[j]
            x_h = np.array(x, dtype=np.float64, copy=True)
            x_h[i] += h[j]

            ## consider bounds
            lower_bound = bounds[i][0]
            upper_bound = bounds[i][1]
            violates_lower_bound = x_h[i] < lower_bound
            violates_upper_bound = x_h[i] > upper_bound

            if accuracy_order == 1:
                if violates_lower_bound or violates_upper_bound:
                    h[j] *= -1
                    x_h[i] = x[i] + h[j]
            else:
                if violates_lower_bound or violates_upper_bound:
                    if violates_lower_bound:
                        x_h[i] = lower_bound
                    else:
                        x_h[i] = upper_bound
            
            ## recalculate h   (improvement of accuracy of h)
            h[j] = x_h[i] - x[i]

            ## eval f and add to df
            f_x_h = f(x_h)
            f_x_h = np.asanyarray(f_x_h)

            if df is None:
                df_shape = (n,) + f_x_h.shape
                df = np.zeros(df_shape)
            
            df[i] += (-1)**j * f_x_h

        ## calculate df_i
        if accuracy_order == 1:
            df[i] -= f_x
            df[i] /= h
        else:
            df[i] /= np.sum(np.abs(h))

    return df



try:
    import util.petsc.with_petsc4py as petsc_module
except ImportError:
    import util.petsc.without_petsc4py as petsc_module



def load_petsc_vec_to_numpy_array(file):
    return petsc_module.load_petsc_vec_to_numpy_array(file)

def save_numpy_array_to_petsc_vec(file, vec):
    import util.petsc.without_petsc4py
    util.petsc.without_petsc4py.save_numpy_array_to_petsc_vec(file, vec)


def load_petsc_mat_to_array(file, dtype=float):
    return petsc_module.load_petsc_mat_to_array(file, dtype=dtype)


import logging
import os
import sys

import util.parallel.is_running


def get_logger_logging():
    return logging.getLogger()

def get_logger_multiprocessing():
    import multiprocessing
    return multiprocessing.get_logger()

def get_logger_scoop():
    import scoop
    return scoop.logger


def get_logger():
    if util.parallel.is_running.scoop_module():
        logger = get_logger_scoop()
        logger.debug('Returning logger from scoop module.')
    elif util.parallel.is_running.multiprocessing_module():
        logger = get_logger_multiprocessing()
        logger.debug('Returning logger from multiprocessing module.')
    else:
        logger = get_logger_logging()
        logger.debug('Returning logger from logging module.')

    return logger


logger = get_logger()


def is_debug():
    return logger.level <= logging.DEBUG


class Logger():

    def __init__(self, disp_stdout=True, log_file=None):
        disp_stdout = disp_stdout or (disp_stdout is None and log_file is None)
        self.enabled = disp_stdout or log_file is not None

        if self.enabled:
            self.log_file = log_file
            self.handlers = []

            ## set logger
            logger.level = logging.DEBUG
            self.logger = logger

            ## create formatter
            formatter_normal = logging.Formatter(
                "%(levelname)-8s %(module)s:%(lineno)d -> %(message)s",
                datefmt=None,
                style='%')
            try:
                import colorlog
            except ImportError:
                colorlog = None
                logger.warn('The package colorlog is not installed. Using uncolored logging.')
            if colorlog is not None:
                formatter_stdout = colorlog.ColoredFormatter(
                    "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(module)s:%(lineno)d -> %(message)s",
                    datefmt=None,
                    reset=True,
                    log_colors={'DEBUG':    'cyan',
                                'INFO':     'green',
                                'WARNING':  'yellow',
                                'ERROR':    'red',
                                'CRITICAL': 'red,bg_white'},
                    secondary_log_colors={},
                    style='%')
            else:
                formatter_stdout = formatter_normal

            ## add stdout and stderr handler
            if disp_stdout:
                if not self.has_stream_handler(sys.stdout) and not self.has_stream_handler(sys.stderr):
                    handler = logging.StreamHandler(sys.stdout)
                    handler.setFormatter(formatter_stdout)
                    self.add_handler(handler)

                    # ## set color formatter
                    # try:
                    #     import colorlog
                    # except ImportError:
                    #     colorlog = None
                    #     logger.warn('The package colorlog is not installed. Using uncolored logging.')
                    # if colorlog is not None:
                    #     formatter = colorlog.ColoredFormatter(
                    #         "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(module)s:%(lineno)d -> %(message)s",
                    #         datefmt=None,
                    #         reset=True,
                    #         log_colors={'DEBUG':    'cyan',
                    #                     'INFO':     'green',
                    #                     'WARNING':  'yellow',
                    #                     'ERROR':    'red',
                    #                     'CRITICAL': 'red,bg_white'},
                    #         secondary_log_colors={},
                    #         style='%')
                    # else:
                    #     formatter = logging.Formatter(
                    #         "%(levelname)-8s %(module)s:%(lineno)d -> %(message)s",
                    #         datefmt=None,
                    #         style='%')
                    # for handler in logger.handlers:
                    #     handler.setFormatter(formatter)

            ## add log file handler
            if log_file is not None and log_file != '':
                try:
                    os.remove(log_file)
                except (OSError, IOError):
                    pass
                handler = logging.FileHandler(log_file)
                handler.setFormatter(formatter_normal)
                self.add_handler(handler)
            ## add null handler if no other handler
            if not logger.hasHandlers():
                self.add_handler(logging.NullHandler())

    #         ## set formatter
    #         formatter = logging.Formatter(fmt="{levelname}:{pathname}:{lineno} ->  {message}", style='{')
    #         for handler in logger.handlers:
    #             handler.setFormatter(formatter)

            ## debug infos
            if disp_stdout:
                logger.debug('Logger {} configured with output to stdout.'.format(logger.name))
            if log_file is not None:
                logger.debug('Logger {} configured with output to file {}.'.format(logger.name, log_file))


#         if disp_stdout:
#             logging.basicConfig(format="{levelname}:{pathname}:{lineno} ->  {message}", style='{', level=logging.DEBUG)
#         else:
#             try:
#                 os.remove(log_file)
#             #except FileNotFoundError:
#             except (OSError, IOError):
#                 pass
#             logging.basicConfig(format="{levelname}:{pathname}:{lineno} ->  {message}", style='{', level=logging.DEBUG, filename=log_file)
#             if disp_stdout:
#                 logging.getLogger().addHandler(logging.StreamHandler(sys.stdout))

    def __del__(self):
        self.close()

    def __enter__(self):
        if self.enabled:
            self.logger.info('Starting logging.')
        return self

    def __exit__(self, eType, eValue, eTrace):
        if self.enabled:
            if eValue is not None and eType is not SystemExit:
                self.logger.exception(eValue)
            self.logger.info('Logging stopped.')
        self.close()


    def close(self):
        if self.enabled:
            self.remove_all_handler()
#         for handler in self.logger.handlers:
#             handler.flush()
#             handler.close()
#         logging.shutdown()

    ## handler

    def add_handler(self, handler):
        self.handlers.append(handler)
        self.logger.addHandler(handler)

    def remove_handler(self, handler):
        handler.flush()
        handler.close()
        self.handlers.remove(handler)
        self.logger.removeHandler(handler)

    def remove_all_handler(self):
        for handler in self.handlers:
            self.remove_handler(handler)

    def has_stream_handler(self, stream):
        for handler in self.logger.handlers:
            try:
                if handler.stream == stream:
                    return True
            except AttributeError:
                pass
        return False




import os

def load(variable_name):
    try:
        return os.environ[variable_name]
    except KeyError:
        raise EnvironmentLookupError(variable_name)


PYTHONPATH_ENV_NAME = 'PYTHONPATH'
def pythonpath():
    try:
        return os.environ[PYTHONPATH_ENV_NAME]
    except KeyError:
        return ''


class EnvironmentLookupError(LookupError):

    def __init__(self, variable_name):
        message = 'Environment variable {} is not set!'.format(variable_name)
        super().__init__(message)



import numpy as np

import util.logging
logger = util.logging.logger


## netcdf

def load_with_scipy(file, data_name):
    import scipy.io

    """
    Loads data from a netcdf file.

    Parameters
    ----------
    file : string or file-like
        The name of the netcdf file to open.
    data_name : string
        The name of the data to extract from the netcdf file.

    Returns
    -------
    data : ndarray
        The desired data from the netcdf file as ndarray with nan for missing values.
    """

    logger.debug('Loading data {} of netcdf file {} with scipy.io.'.format(data_name, file))

    f = scipy.io.netcdf.netcdf_file(file, 'r')
    data_netcdf = f.variables[data_name]
    data = np.array(data_netcdf.data, copy = True)
    data[data == data_netcdf.missing_value] = np.nan
    f.close()

    return data


def load_with_netcdf4(file, data_name):
    import netCDF4

    logger.debug('Loading data {} of netcdf file {} with netCDF4.'.format(data_name, file))

    nc_file = netCDF4.Dataset(file, 'r')
    data = nc_file.variables[data_name][:]
    nc_file.close()

    return data


def load(file, data_name):
    try:
        import netCDF4
    except:
        pass
    else:
        return load_with_netcdf4(file, data_name)
    
    try:
        import scipy.io
    except:
        pass
    else:
        return load_with_scipy(file, data_name)
    
    raise ImportError('Could not import netCDF4 nor scipy.io'.)
    


import numpy as np

import logging
logger = logging.getLogger(__name__)


PETSC_VEC_HEADER = 1211214

def load_petsc_vec_to_numpy_array(file):
    logger.debug('Loading petsc vector from {} to numpy vector.'.format(file))

    with open(file, "rb") as file_object:
        ## omit header
        np.fromfile(file_object, dtype='>i4', count=1)
        ## read length
        nvec = np.fromfile(file_object, dtype='>i4', count=1)
        assert nvec.ndim == 1 and len(nvec) == 1
        nvec = nvec[0]
        ## read values
        v = np.fromfile(file_object, dtype='>f8', count=nvec)
        assert v.ndim == 1 and len(v) == nvec

    return v


def save_numpy_array_to_petsc_vec(file, vec):
    logger.debug('Saving numpy vector to petsc vector in {}.'.format(file))

    with open(file, mode='xb') as file_object:
        ## write header
        header = np.array(PETSC_VEC_HEADER, dtype='>i4')
        header.tofile(file_object)

        ## write length (32 bit int)
        length = np.array(len(vec), dtype='>i4')
        length.tofile(file_object)

        ## write values
        vec = vec.astype('>f8')
        vec.tofile(file_object)




def load_petsc_mat_to_array(file, dtype=float):
    logger.debug('Loading petsc matrix from %s.', file)

    ## open file
    f = open(file, "rb")

    ## omit header
    np.fromfile(f, dtype=">i4", count=1)
    ## read dims
    nx     = np.fromfile(f, dtype=">i4", count=1)
    ny     = np.fromfile(f, dtype=">i4", count=1)
    nnz    = np.fromfile(f, dtype=">i4", count=1)
    nrow   = np.fromfile(f, dtype=">i4", count=nx)
    colidx = np.fromfile(f, dtype=">i4", count=nnz)
    val    = np.fromfile(f, dtype=">f8", count=nnz)

    ## close file
    f.close()

    ## create full matrix
    array = np.zeros(shape=(nx, ny), dtype=dtype)
    offset = 0
    for i in range(nx):
        if not nrow[i] == 0.0:
            for j in range(nrow[i]):
                array[i, colidx[offset]] = dtype(val[offset])
                offset = offset + 1

    ## return array
    return array


import warnings
import numpy as np
# import itertools

import util.logging
logger = util.logging.logger

from .constants import MODES

CURRENT_MODE = MODES['scoop']


def max_parallel_mode():
    if util.parallel.is_running.scoop_module():
        max_parallel_mode = MODES['scoop']
    else:
        max_parallel_mode = MODES['multiprocessing']
    logger.debug('Maximal parallel mode is {}.'.format(max_parallel_mode))
    return max_parallel_mode




## map functions

# def map_serial_with_args(function, values, *args):
#     logger.debug('Mapping function with {} args of types {} to values in serial.'.format(len(args), tuple(map(type, args))))
#
#     values = args_generator_with_indices(values, args)
#     results = itertools.starmap(function, values)
#     results = tuple(results)
#
#     logger.debug('Serial calculation with {} results completed.'.format(len(results)))
#
#     return results


# def map_parallel(function, values, parallel_mode=MODES['scoop']):
#     assert callable(function)
#
#     if parallel_mode == MODES['scoop']:
#         CURRENT_MODE = MODES['multiprocessing']
#         import util.parallel.with_scoop
#         results = util.parallel.with_scoop.map_parallel(function, values)
#     elif parallel_mode == MODES['multiprocessing']:
#         CURRENT_MODE = MODES['serial']
#         import util.parallel.with_multiprocessing
#         results = util.parallel.with_multiprocessing.map_parallel(function, values)
#     else:
#         logger.debug('Calculating in serial.')
#         results = map(function, values)
#
#     return results


def map_parallel_with_args(function, values, *args, parallel_mode=MODES['scoop'], number_of_processes=None, chunksize=1, share_args=True):
    assert callable(function)

    if parallel_mode == MODES['scoop']:
        CURRENT_MODE = MODES['multiprocessing']
        import util.parallel.with_scoop
        results = util.parallel.with_scoop.map_parallel_with_args(function, values, *args)
    elif parallel_mode == MODES['multiprocessing']:
        CURRENT_MODE = MODES['serial']
        import util.parallel.with_multiprocessing
        results = util.parallel.with_multiprocessing.map_parallel_with_args(function, values, *args, number_of_processes=number_of_processes, chunksize=chunksize, share_args=share_args)
    else:
        import util.parallel.with_serial
        results = util.parallel.with_serial.map_serial_with_args(function, values, *args)

    return results




# def create_array(shape, function, args=None, parallel_mode=MODES['scoop']):
#     logger.debug('Creating array with shape {} with parallel mode {}.'.format(shape, parallel_mode))
#
#     ## check that mode is not in use
#     if parallel_mode > CURRENT_MODE:
#         logger.debug('Parallel mode {} used before. Switching to parallel mode {}.'.format(parallel_mode, CURRENT_MODE))
#         parallel_mode = CURRENT_MODE
#
#     ## check if scoop running otherwise fallback
#     if parallel_mode == MODES['scoop'] and not util.parallel.is_running.scoop_module():
#         parallel_mode = MODES['multiprocessing']
#         logger.debug('Scoop is not running falling back to parallel mode {}.'.format(parallel_mode))
#
# #     ## if multiprocessing share arrays
# #     if parallel_mode == MODES['multiprocessing']:
# #         args = util.parallel.with_multiprocessing.share_all_arrays(args, lock=False)
#
#     ## create args generator
#     values = args_generator(shape, args)
#
#     ## execute in parallel
#     results = map_parallel(function, values, parallel_mode=parallel_mode)
#
#     ## create array
#     array = np.array(list(results))
#     logger.debug('Calculation completed. Got {} results.'.format(len(array)))
#     array = array.reshape(shape)
#     return array


## create array

def create_array(shape, function, *args, parallel_mode=MODES['scoop'], number_of_processes=None, chunksize=1, share_args=True):
    logger.debug('Creating array with shape {} with parallel mode {}.'.format(shape, parallel_mode))

    ## check that mode is not in use
    if parallel_mode > CURRENT_MODE:
        logger.debug('Parallel mode {} used before. Switching to parallel mode {}.'.format(parallel_mode, CURRENT_MODE))
        parallel_mode = CURRENT_MODE

    ## check if scoop running otherwise fallback
    if parallel_mode == MODES['scoop'] and not util.parallel.is_running.scoop_module():
        parallel_mode = MODES['multiprocessing']
        logger.debug('Scoop is not running falling back to parallel mode {}.'.format(parallel_mode))

    ## execute in parallel
    results = map_parallel_with_args(function, np.ndindex(*shape), *args, parallel_mode=parallel_mode, number_of_processes=number_of_processes, chunksize=chunksize, share_args=share_args)

    ## create array
    array = np.array(tuple(results))
#     logger.debug('Calculation completed. Got {} results.'.format(len(array)))
    array = array.reshape(shape)
    return array



## args generators

def args_generator_with_shape(shape, args, index_position=0):
    logger.debug('Creating arg generator for shape {} and {} args with index position {}.'.format(shape, len(args), index_position))
    indices = np.ndindex(*shape)
    args_generator_with_indices(indices, args, index_position=index_position)


def args_generator_with_indices(indices, args, index_position=0):
    logger.debug('Creating arg generator for indices and {} args of types {} with index position {}.'.format(len(args), tuple(map(type, args)), index_position))

    ## make list of function args
    if args is not None:
        args_list = list(args)
    else:
        args_list = []

    ## insert index
    args_list.insert(index_position, None)

    ## return function to compute function args with current index
    for index in indices:
        args_list[index_position] = index
#         logger.debug('Returning index {} together with {} args.'.format(index, len(args)))
        yield tuple(args_list)




## zipped function
def eval_function_with_index_and_args(args):
    assert len(args) >= 2
    index = args[0]
    function = args[1]
    args = args[2:]
    return function(index, *args)


#
#
# def map_parallel(function, values, parallel=True):
#     assert callable(function)
#
# #     if parallel:
# #         try:
# #             results = util.parallel.with_scoop.map_parallel(function, values)
# #             scoop_error = False
# #         except RuntimeWarning:
# #             logger.debug('Scoop is not loaded as module.')
# #             scoop_error = True
# #
# #     if not parallel or scoop_error:
# #         logger.debug('Calculating in serial.')
# #         results = map(function, values)
# #         logger.debug('Serial calculation completed.')
#
#     if parallel:
#         results = util.parallel.with_scoop.map_parallel(function, values)
#     else:
#         logger.debug('Calculating in serial.')
#         results = map(function, values)
#
#     return results
#
#
#
# def create_array(shape, function, args=None, index_position=0, parallel=True):
#     logger.debug('Creating array with shape {} in parallel {}.'.format(shape, parallel))
#
#     ## check if scoop running
#     if parallel and not util.parallel.is_running.scoop_module():
#         logger.debug('Scoop is not running falling back to serial mode.')
#         parallel = False
#
#     ## execute in parallel
#     results = map_parallel(function, args_function_generator(shape, args, index_position), parallel=parallel)
#
#     ## create array
#     array = np.array(list(results))
#     logger.debug('Calculation completed. Got {} results.'.format(len(array)))
#     array = array.reshape(shape)
#     return array
#
#
# # def create_array(shape, function, args=None, constants=True, index_position=0, parallel=True):
# #     logger.debug('Creating array with shape {} in parallel {}.'.format(shape, parallel))
# #
# #     ## check if scoop running
# #     if parallel and not util.parallel.is_running.scoop_module():
# #         logger.debug('Scoop is not running falling back to serial mode.')
# #         parallel = False
# #
# #
# #     def args_function_generator(shape, args, constants, index_position, parallel):
# #         logger.debug('Create arg function generator.')
# #
# #         ## make list of function args
# #         if args is not None:
# #             args_list = list(args)
# #         else:
# #             args_list = []
# #
# #         ## create compute function arg function
# #         if parallel and constants:
# #             ## set constants
# #             constants_uuids = [util.parallel.with_scoop.set_shared_object(arg) for arg in args]
# #
# #             def args_function(index):
# #                 args_list = [util.parallel.with_scoop.get_shared_object(constants_uuid) for constants_uuid in constants_uuids]
# #                 args_list.insert(index_position, index)
# #                 return tuple(args_list)
# #         else:
# #             args_list.insert(index_position, None)
# #             def args_function(index):
# #                 args_list[index_position] = index
# #                 return tuple(args_list)
# #
# #         logger.debug('Arg function generator created.')
# #
# #         ## return function to compute function args with current index
# #         indices = np.ndindex(*shape)
# #         for index in indices:
# #             yield lambda :args_function(index)
# #
# #     ## execute in parallel
# #     results = map_parallel(function, args_function_generator(shape, args, constants, index_position, parallel), parallel=parallel)
# #
# # #         ## add index to function args
# # #         list_function_args.insert(index_position, None)
# # #
# # #         ## return function args with current index
# # #         indices = np.ndindex(*shape)
# # #         for index in indices:
# # #             list_function_args[index_position] = index
# # #             yield tuple(list_function_args)
# # #
# # #
# # #     ## execute in parallel
# # #     results = map_parallel(function, function_arg_generator(shape, function_args, index_position), parallel=parallel)
# #
# #     ## create array
# #     array = np.array(list(results))
# #     logger.debug('Got {} results in parallel.'.format(len(array)))
# #     array = array.reshape(shape)
# #     return array


import re


def get_int_in_string(string):
    return int(re.search(r'\d+', string).group())


def convert_format_string_in_regular_expression(format_string):
    regular_expression = format_string
    for pattern in re.findall('{.*:.*d}', regular_expression):
        regular_expression = regular_expression.replace(pattern, '[+,-]?[0-9]*')
    for pattern in re.findall('{.*}', regular_expression):
        regular_expression = regular_expression.replace(pattern, '.*')
    return regular_expression



QUEUES = ('clexpress', 'clmedium', 'cllong', 'long', 'clbigmem', 'clfocean', 'clfo2', 'feque')
MAX_WALLTIME = {'clexpress':2, 'clmedium':48, 'cllong':100, 'clbigmem':100, 'clfocean':100, 'clfo2':100, 'feque':10}

MODEL_RENAMING = {'python':'Python3.3.6', 'python3':'Python3.3.6', 'hdf5':'hdf5-1.8.15', 'matlab':'matlab2015a', 'petsc':'petsc-3.3-p4-intel', 'intel14': 'intel', 'interlmpi14': 'intelmpi', 'intel16': 'intel16.0.3', 'intelmpi16': 'intelmpi16.0.3'}

COMMANDS = {'mpirun': 'mpirun $NQSII_MPIOPTS -np {cpus:d} {command}',
            'time': 'TIME_FMT="\nStatistics for %C:\nElapsed time: %Es, Exit code: %x\nCPU: %Us user mode, %Ss kernel mode, %P workload\nMemory: %Mkb max, %W swap outs\nContex switches: %c involuntarily, %w voluntarily\nPage faults: %F major, %R minor\nFile system I/O: %I inputs, %O outputs"\n/sfs/fs3/sw/tools/time1.7/bin/time -f "$TIME_FMT" {command}',
            'sub': '/usr/bin/nqsII/qsub',
            # 'stat': '/usr/bin/nqsII/qstat',
            'stat': '/usr/local/bin/qstatall',
            'nodes': '/sfs/fs3/sw/tools/qcl/qcl',
            'python': '/sfs/fs5/home-sh/sunip229/local/bin/python3'}

NODE_INFOS = {'clexpress': {'nodes': 2, 'speed': 2.6, 'cpus': 16, 'memory': 128, 'max_walltime': MAX_WALLTIME['clexpress']},
              'clmedium': {'nodes': 60, 'speed': 2.6, 'cpus': 16, 'memory': 128, 'max_walltime': MAX_WALLTIME['clmedium']},
              'cllong': {'nodes': 30, 'speed': 2.6, 'cpus': 16, 'memory': 128, 'max_walltime': MAX_WALLTIME['cllong']},
              'clbigmem': {'nodes': 4, 'speed': 2.6, 'cpus': 16, 'memory': 256, 'max_walltime': MAX_WALLTIME['clbigmem']},
              'clfocean': {'nodes': 4, 'speed': 2.6, 'cpus': 16, 'memory': 128, 'max_walltime': MAX_WALLTIME['clfocean']},
              'clfo2': {'nodes': 18, 'speed': 2.5, 'cpus': 24, 'memory': 128, 'max_walltime': MAX_WALLTIME['clfo2']},
              'feque': {'nodes': 1, 'speed': 0, 'cpus': 16, 'memory': 128, 'leave_free': 1, 'max_walltime': MAX_WALLTIME['feque']}}


EXCEEDED_WALLTIME_ERROR_MESSAGE = "Batch job received signal SIGKILL. (Exceeded per-req elapse time limit)"


import numpy as np

import util.io.fs


FILE_EXT = '.{matrix_format}.npz'
SPARSE_FORMATS= ('csc', 'csr', 'bsr')


def is_spmatrix(obj):
    try:
        import scipy.sparse
    except ImportError:
        return False
    else:
        return scipy.sparse.isspmatrix(obj)

def is_file(file):
    return np.any([util.io.fs.has_file_ext(file, FILE_EXT.format(matrix_format=matrix_format)) for matrix_format in SPARSE_FORMATS])

def add_file_ext(file, sparse_format):
    file_ext = FILE_EXT.format(matrix_format=sparse_format)
    return util.io.fs.add_file_ext_if_needed(file, file_ext)


def save(file, matrix):
    import scipy.sparse
    file = add_file_ext(file, sparse_format=matrix.format)
    scipy.sparse.save_npz(file, matrix)

def load(file):
    import scipy.sparse
    return scipy.sparse.load_npz(file)




import itertools

import util.parallel.universal

import util.logging
logger = util.logging.logger



def map_serial_with_args(function, indices, *args):
    logger.debug('Mapping function with {} args of types {} to values in serial.'.format(len(args), tuple(map(type, args))))

    values = util.parallel.universal.args_generator_with_indices(indices, args)
    results = itertools.starmap(function, values)
    results = tuple(results)

    logger.debug('Serial calculation with {} results completed.'.format(len(results)))

    return results



import os

import util.io.np
import util.io.object
import util.io.matrix.sparse


def save(file, o, make_dirs=True):
    if make_dirs:
        dir = os.path.dirname(file)
        if len(dir) > 0:
            os.makedirs(dir, exist_ok=True)
    
    if util.io.matrix.sparse.is_spmatrix(o):
        util.io.matrix.sparse.save(file, o)
    elif util.io.np.is_file(file):
        util.io.np.save(file, o)
    else:
        util.io.object.save(file, o)


def load(file):
    if util.io.matrix.sparse.is_file(file):
        return util.io.matrix.sparse.load(file)
    elif util.io.np.is_file(file):
        return util.io.np.load(file)
    else:
        return util.io.object.load(file)


import os.path
import subprocess

import numpy as np

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.colorbar
import matplotlib.colors
import matplotlib.font_manager
import matplotlib.text
import matplotlib.ticker
import mpl_toolkits.axes_grid1

import util.io.fs
import util.logging
logger = util.logging.logger

DEFAULT_FONT_SIZE = 20


## plot types

def data(data, file, land_value=np.nan, no_data_value=np.inf, land_brightness=0, use_log_scale=False, v_min=None, v_max=None, caption=None, tick_font_size=DEFAULT_FONT_SIZE, power_limit=3, dpi=100, contours=False, contours_text_brightness=0.5, colorbar=True):
    
    if data.dtype == np.float128:
        data = data.astype(np.float64)
    

    def get_masks(data, land_value=np.nan, no_data_value=0):
        def get_value_mask(array, value):
            if value is None:
                mask = np.zeros_like(array).astype('bool')
            elif value is np.nan:
                mask = np.isnan(array)
            else:
                mask = array == value
            return mask


        land_mask = get_value_mask(data, land_value)
        no_data_mask = get_value_mask(data, no_data_value)

        return (land_mask, no_data_mask)


    logger.debug('Plotting data.')

    ## set font size
    set_global_font_size(tick_font_size)


    ## reshape data
    original_shape = data.shape
    original_dim = len(original_shape)
    if original_dim == 2:
        data = data.reshape((1,) + original_shape + (1,))
    elif original_dim == 3:
        data = data.reshape((1,) + original_shape)


    ## get masks and v_min and v_max
    (land_mask, no_data_mask) = get_masks(data, land_value=land_value, no_data_value=no_data_value)
    if v_min is None or v_max is None or contours:
        data_mask = np.logical_not(np.logical_or(no_data_mask, land_mask))
        data_min = np.min(data[data_mask])
        data_max = np.max(data[data_mask])

    ## v_min and v_max
    if v_min is None:
        v_min = data_min
    if v_max is None:
        v_max = data_max
    logger.debug('Using {} as v_min and {} as v_max.'.format(v_min, v_max))


    ## remove negative values for log plot
    if use_log_scale:
        data = np.array(data, copy=True)
        data[land_mask] = np.nan
        land_value = np.nan
        data[no_data_mask] = np.inf
        no_data_value = np.inf
        if v_min <= 1:
            v_min = 1


    ## splite filename
    file_root, file_extension = os.path.splitext(file)


    ## prepare no_data_array
    no_data_array = np.empty_like(data[0,:,:,0])

    t_len = data.shape[0]
    t_len_str = str(t_len)
    z_len = data.shape[3]
    z_len_str = str(z_len)

    for z in range(z_len):
        current_file_with_z = file_root

        ## append depth to filename
        if z_len > 1:
            current_file_with_z += '_layer_' + z_len_str + '_' + str(z+1).zfill(len(z_len_str))

        for t in range(t_len):
            current_file = current_file_with_z

            ## append time to filename
            if t_len > 1:
                current_file += '_time_' + t_len_str + '_' + str(t+1).zfill(len(t_len_str))

            current_file += file_extension

            ## make no_data with 1 where no data, 0.5 where water at surface, 0 where land and nan where data (1 is white, 0 is black)
            no_data_array = no_data_array * np.nan
            no_data_array[no_data_mask[t,:,:,z]] = 1
            no_data_array[land_mask[t,:,:,z]] = (1 + 9 * land_brightness) / 10
            no_data_array[land_mask[t,:,:,0]] = land_brightness
            

            ## make figure
            fig = plt.figure()

            ## chose colormap
            colormap = plt.cm.jet
            colormap.set_bad(color='w', alpha=0.0)

            ## chose norm
            if use_log_scale:
                norm = matplotlib.colors.LogNorm(vmin=v_min, vmax=v_max)
            else:
                if issubclass(data.dtype.type, np.integer):
                    norm = matplotlib.colors.BoundaryNorm(np.arange(v_min, v_max+1), colormap.N)
                else:
                    norm = None
                

            ## plot no data mask
            colormap_no_data = plt.cm.gray
            colormap_no_data.set_bad(color='w', alpha=0.0)
            axes_image = plt.imshow(no_data_array.transpose(), origin='lower', aspect='equal', cmap=colormap_no_data, vmin=0, vmax=1)

            ## plot data
            current_data = data[t,:,:,z].transpose()
            axes_image = plt.imshow(current_data, origin='lower', aspect='equal', cmap=colormap, vmin=v_min, vmax=v_max, norm=norm)

            ## disable axis labels
            plt.axis('off')
            
            ## chose tick locator
            if colorbar or contours:
                ## chose tick base
                if colorbar:
                    v_max_tick = v_max
                else:
                    current_data_max = np.nanmax(current_data)
                    v_max_tick = min([v_max, current_data_max])
                    if current_data_max == 0:
                        contours = False
                        v_max_tick = 1
                tick_base_exp = int(np.ceil(np.log10(v_max_tick))) - 1
                tick_base = 10 ** tick_base_exp
                
                if (current_data > tick_base).sum() < (~np.isnan(current_data)).sum() / 100:    # decrease tick if too few data above tick
                    tick_base = tick_base / 10
                
                ## chose locator
                if use_log_scale:
                    tick_locator = plt.LogLocator(base=tick_base, subs=tick_base/10)
                else:
                    tick_locator = plt.MultipleLocator(base=tick_base)

            ## plot colorbar
            if colorbar:
                divider = mpl_toolkits.axes_grid1.make_axes_locatable(plt.gca())
                cax = divider.append_axes("right", size="3%", pad=0.1) 
                               
                cb = plt.colorbar(axes_image, cax=cax)
                cb.locator = tick_locator
                cb.update_ticks()
                
                if not use_log_scale:
                    cb.formatter.set_powerlimits((-power_limit, power_limit))
                    cb.update_ticks()
            
            ## plot contours
            if contours:
                contour_plot = plt.contour(current_data, locator=tick_locator, colors='k', linestyles=['dashed', 'solid'], linewidths=0.5, norm=norm)
                if np.abs(tick_base_exp) >= power_limit:
                    label_fmt = '%.0e'
                elif tick_base_exp < 0:
                    label_fmt = '%1.{:d}f'.format(np.abs(tick_base_exp))
                else:
                    label_fmt = '%d'
                plt.clabel(contour_plot, contour_plot.levels[1::2], fontsize=8, fmt=label_fmt, colors=str(contours_text_brightness))
            
    
            ## set caption
            if caption is not None:
                plt.xlabel(caption, fontsize=font_size, fontweight='bold')

            ## save and close
            save_and_close_fig(fig, current_file, dpi=dpi)

    logger.debug('Plot completed.')





def line(x, y, file, x_order=0, line_label=None, line_width=1, line_style='-', line_color='r', y_min=None, y_max=None, xticks=None, spine_line_width=1, use_log_scale=False, transparent=True, tick_font_size=DEFAULT_FONT_SIZE, legend_font_size=DEFAULT_FONT_SIZE, x_label=None, y_label=None, axis_label_font_size=DEFAULT_FONT_SIZE, dpi=800):
    logger.debug('Plotting line.')
    
    ## check if multi line
    try:
        y[0][0]
    except LookupError:
        is_multi_line = False
    else:
        is_multi_line = True
    
    ## check input and prepare
    if len(x) != len(y):
        if not is_multi_line:
                ValueError('For single line plot, x and y must have same length but length x is {} and length y is {}.'.format(len(x), len(y)))
        else:
            x = np.asarray(x)
            y = np.asarray(y)
            assert y.ndim == 2
            if x.ndim != 1:
                raise ValueError('For multiline plot, if x and y have two dims, the x and y must have same length but length x is {} and length y is {}.'.format(len(x), len(y)))
            else:
                if x.shape[0] != y.shape[1]:
                    raise ValueError('For multiline plot, if x has one dim and y has two dims, the first dim of x must have the same size as the second dim of y, but shape x is {} and shape y is {}.'.format(x.shape, y.shape))
                else:
                    x = np.tile(x, y.shape[1]).reshape(y.shape[1], -1)
        

    ## prepare line(s)
    if len(y) > 0:
        ## multiple lines
        if is_multi_line:
            xs = x
            ys = y
    
            ## copy line setups for each line if only one line setup passed
            number_of_lines = len(ys)
            line_setups = []
            for line_setup in (line_label, line_width, line_style, line_color):
                if np.asarray(line_setup).ndim == 0:
                    line_setup = (line_setup,) * len(y)
                line_setups.append(line_setup)
    
            line_labels = line_setups[0]
            line_widths = line_setups[1]
            line_styles = line_setups[2]
            line_colors = line_setups[3]
        ## one line
        else:
            number_of_lines = 1
            xs = x.reshape((1,) + x.shape)
            ys = y.reshape((1,) + y.shape)
            line_labels = [line_label]
            line_widths = [line_width]
            line_styles = [line_style]
            line_colors = [line_color]
    else:
        number_of_lines = 0

    ## make figure
    fig = plt.figure()
    x_min = np.inf
    x_max = -np.inf

    ## plot each line
    for i in range(number_of_lines):
        ## get line and setup
        x = np.asanyarray(xs[i])
        y = np.asanyarray(ys[i])
        line_label = line_labels[i]

        if len(x) > 0:
            if line_label is None:
                line_label = '_nolegend_'
            line_width = line_widths[i]
            line_style = line_styles[i]
            line_color = line_colors[i]

            ## sort values
            if x_order != 0:
                sorted_indices = np.argsort(x)
                if x_order < 0:
                    sorted_indices = sorted_indices[::-1]
                x = x[sorted_indices]
                y = y[sorted_indices]

            ## update x_min x_max
            x_min = min([x_min, x[0]])
            x_max = max([x_max, x[-1]])

            ## plot line
            plt.plot(x, y, line_style, color=line_color, linewidth=line_width, markersize=line_width*3, label=line_label)


    ## set axis labels
    if x_label is not None:
        plt.xlabel(x_label, fontsize=axis_label_font_size, fontweight='bold')
    if y_label is not None:
        plt.ylabel(y_label, fontsize=axis_label_font_size, fontweight='bold')

    ## set tickts
    if xticks is not None:
        plt.xticks(xticks)
    set_tick_font(fig.gca(), size=tick_font_size)

    ## set spine line_width
    set_spine_line_size(fig, spine_line_width)

    ## log scale
    if use_log_scale:
        plt.yscale('log')

    ## set axis limits
    if y_min is not None:
        plt.ylim(ymin=y_min)
    if y_max is not None:
        plt.ylim(ymax=y_max)
    plt.xlim(x_min, x_max)

    ## legend
    legend = plt.legend(loc=0)
    if legend is not None:
        legend.get_frame().set_alpha(float(not transparent))
        set_legend_font(fig, size=legend_font_size)

    ## save and close
    save_and_close_fig(fig, file, dpi=dpi)



def scatter(x, y, file, point_size=20, dpi=800):
    ## check and prepare input
    if x.ndim == 2 and x.shape[1] > 2:
        raise ValueError('Scatter plots for x dim {} is not supported.'.format(x.shape[1]))
    if x.ndim == 2 and x.shape[1] == 1:
        x = x[:,0]

    ## make figure
    fig = plt.figure()

    ## plot
    if x.ndim == 1:
        plt.scatter(x, y, s=point_size)
    if x.ndim == 2:
        ax = fig.add_subplot(111, projection='3d')
        ax.scatter(x[:,0], x[:,1], y, s=point_size)

    ## save and close
    save_and_close_fig(fig, file, dpi=dpi)



def histogram(data, file, bins=None, step_size=None, x_min=None, x_max=None, weights=None, use_log_scale=False, type='bar', tick_font_size=DEFAULT_FONT_SIZE, tick_power=None, dpi=800):
    logger.debug('Plotting histogram.')

    ## make fig
    fig = plt.figure()

    ##
    if bins is None:
        if step_size is None:
            raise ValueError('Either "bins" or "step_size" has to be defined.')
        if x_min is None:
            x_min = np.floor(np.min(data) / step_size) * step_size
        if x_max is None:
            x_max = np.ceil(np.max(data) / step_size) * step_size
        bins = np.arange(x_min, x_max+step_size, step_size)

    ## plot
    (n, bins, patches) = plt.hist(data, bins=bins, weights=weights, log=use_log_scale, histtype=type)
    plt.xlim(bins[0], bins[-1])

    ## set axis label size
    if tick_power is not None:
        set_tick_power_fix(fig.gca(), axis='y', power=tick_power)
    set_tick_font(fig.gca(), size=tick_font_size)

    ## save and close
    save_and_close_fig(fig, file, dpi=dpi)



def spy(A, file, markersize=1, axis_labels=True, caption=None, font_size=DEFAULT_FONT_SIZE, dpi=800):
    import scipy.sparse

    ## set font size
    set_global_font_size(font_size)

    ## make figure
    fig = plt.figure()
    
    ## plot sparsity_pattern
    if scipy.sparse.issparse(A):
        logger.debug('Plotting sparsity pattern for matrix {!r} with markersize {} and dpi {} to file {}.'.format(A, markersize, dpi, file))
        plt.spy(A, markersize=markersize, marker=',', markeredgecolor='k', markerfacecolor='k')
    
    ## plot matrix values
    else:
        logger.debug('Plotting values for matrix {!r} with dpi {} to file {}.'.format(A, dpi, file))
        v_abs_max = np.abs(A).max()
        axes_image = plt.imshow(A, cmap=plt.cm.bwr, interpolation='nearest', vmin=-v_abs_max, vmax=v_abs_max)   
        cb = fig.colorbar(axes_image)
        cb.ax.tick_params(labelsize=font_size) 
        fig.gca().set_xticks([]) 
        fig.gca().set_yticks([]) 

    ## set power limits
    if axis_labels:
        formatter = plt.ScalarFormatter()
        formatter.set_powerlimits((-3,3))
        fig.gca().xaxis.set_major_formatter(formatter)
        fig.gca().yaxis.set_major_formatter(formatter)
    
    ## disable axis labels
    else:
        plt.axis('off')
    
    ## set caption
    if caption is not None:
        plt.xlabel(caption, fontsize=font_size, fontweight='bold')
    
    ## save and close
    save_and_close_fig(fig, file, dpi=dpi)




def intervals(intervals, file, use_percent_ticks=False, caption=None, font_size=DEFAULT_FONT_SIZE, dpi=800):
    intervals = np.asanyarray(intervals)
    assert intervals.ndim == 2
    assert len(intervals) == 2
    
    ## calculate data
    intervals_half_sizes = (intervals[1] - intervals[0]) / 2
    means = intervals_half_sizes + intervals[0]
    n = len(intervals_half_sizes)

    ## set font size
    set_global_font_size(font_size)

    ## make figure
    fig = plt.figure()
    
    ## plot
    linewidth = 3
    plt.errorbar(np.arange(1, n+1), means, xerr=0, yerr=intervals_half_sizes, linestyle='', linewidth=linewidth, elinewidth=linewidth, capsize=linewidth*3, capthick=linewidth, marker='.', markersize=linewidth*5)
    
    ## set y limits
    plt.xlim(0.5, n + 0.5)
    
    ## y tick formatter
    if use_percent_ticks:
        fmt = '%.0f%%'
        yticks = matplotlib.ticker.FormatStrFormatter(fmt)
        fig.gca().yaxis.set_major_formatter(yticks)

    ## set caption
    if caption is not None:
        plt.xlabel(caption, fontsize=font_size, fontweight='bold')
    
    ## save and close
    save_and_close_fig(fig, file, dpi=dpi)




## auxiliary functions

def trim(file):
    logger.debug('Trimming plot %s.' % file)
    subprocess.check_output(('convert', '-trim', file, file))


def save_and_close_fig(fig, file, transparent=True, dpi=800, make_read_only=True):
    plt.tight_layout()
    plt.savefig(file, bbox_inches='tight', transparent=transparent, dpi=dpi)
    plt.close(fig)
    trim(file)
    if make_read_only:
        util.io.fs.make_read_only(file)
    logger.debug('Plot saved to {}.'.format(file))



def set_spine_line_size(fig, line_width):
    axes = fig.gca()
    for axis in ['top','bottom','left','right']:
        axes.spines[axis].set_linewidth(line_width)


def set_font_all(fig, family='sans-serif', weight='bold', size=12):
    font_properties = matplotlib.font_manager.FontProperties(family=family, weight=weight, size=size)

    for t in fig.findobj(matplotlib.text.Text):
        t.set_fontproperties(font_properties)



def set_tick_font(axes, family='sans-serif', weight='bold', size=12):
    font_properties = matplotlib.font_manager.FontProperties(family=family, weight=weight, size=size)
    plt.setp(axes.get_xticklabels(), fontproperties=font_properties)
    plt.setp(axes.get_xaxis().get_offset_text(), fontproperties=font_properties)
    plt.setp(axes.get_yticklabels(), fontproperties=font_properties)
    plt.setp(axes.get_yaxis().get_offset_text(), fontproperties=font_properties)


def set_tick_power_limits(axis='y', power_limits=(-2,2)):
    plt.ticklabel_format(style='sci', axis=axis, scilimits=power_limits)

def set_tick_power_fix(axes, axis='y', power=3):
    f = FixedOrderFormatter(power)
    if axis == 'y' or axis == 'both':
        axes.yaxis.set_major_formatter(f)
    if axis == 'x' or axis == 'both':
        axes.xaxis.set_major_formatter(f)

class FixedOrderFormatter(matplotlib.ticker.ScalarFormatter):
    """Formats axis ticks using scientific notation with a constant order of
    magnitude"""
    def __init__(self, order_of_magnitude=0, use_offset=True, use_math_text=None):
        self._order_of_magnitude = order_of_magnitude
        matplotlib.ticker.ScalarFormatter.__init__(self, useOffset=use_offset, useMathText=use_math_text)
        self.orderOfMagnitude = order_of_magnitude

    def _set_orderOfMagnitude(self, range):
        """Over-riding this to avoid having orderOfMagnitude reset elsewhere"""
        self.orderOfMagnitude = self._order_of_magnitude


def set_legend_font(fig, family='sans-serif', weight='bold', size=12):
    font_properties = matplotlib.font_manager.FontProperties(family=family, weight=weight, size=size)
    axes = fig.gca()
    legend = axes.get_legend()
    plt.setp(legend.get_texts(), fontproperties=font_properties)


def set_global_font_size(size=20):
    logger.debug('Setting font size for plots to {}.'.format(size))

    font = {'family' : 'sans-serif',
            'weight' : 'bold',
            'size'   : size
    }
    matplotlib.rc('font', **font)


def get_colors(n, colormap_name='gist_rainbow'):
    colormap = plt.get_cmap(colormap_name)
    colors = [colormap(i/(n-1)) for i in range(n)]
    return colors




import os

import numpy as np

import util.io.fs


FILE_EXT = '.npy'
COMPRESSED_FILE_EXT = '.npz'


def get_ext(compressed=False):
    if compressed:
        return COMPRESSED_FILE_EXT
    else:
        return FILE_EXT


def is_file(file, compressed=None):
    if compressed is None:
        return is_file(file, compressed=False) or is_file(file, compressed=True)
    else:
        ext = get_ext(compressed=compressed)
        return util.io.fs.has_file_ext(file, ext)


def add_file_ext(file, compressed=False):
    ext = get_ext(compressed=compressed)
    return util.io.fs.add_file_ext_if_needed(file, ext)


def save(file, values, compressed=None, make_read_only=False, overwrite=False, create_path_if_not_exists=True):
    ## check input values
    is_values_dict = isinstance(values, dict)
    is_values_tuple = (isinstance(values, tuple) or isinstance(values, list)) and all(map(lambda a: isinstance(a, np.ndarray), values))
    
    if not is_values_dict and not is_values_tuple:
        values = np.asanyarray(values)
    
    if is_file(file, compressed=False):
        if compressed:
            raise ValueError('Compressed values can only be stored in "npz" file format, but the file {} has ending "npy".'.format(file))
        if is_values_dict or is_values_tuple:
            raise ValueError('Multiple values {} can only be stored in "npz" file format, but the file {} has ending "npy".'.format(file))
    
    ## set file ext
    use_npz = is_values_dict or is_values_tuple or (compressed is not None and compressed) or is_file(file, compressed=True)    
    file = add_file_ext(file, compressed=use_npz)
    
    ## set compressed if not passed
    if compressed is None:
        compressed = use_npz
    
    ## create dir
    if create_path_if_not_exists:
        (dir, filename) = os.path.split(file)
        os.makedirs(dir, exist_ok=True)
    
    ## remove if overwrite
    if overwrite:
        util.io.fs.remove_file(file, force=True, not_exist_okay=True)
    
    ## save
    if use_npz:
        if compressed:
            np.savez_compressed(file, values)
        else:
            np.savez(file, values)
    else:
        np.save(file, values)
    
    ## make read only
    if make_read_only:
        util.io.fs.make_read_only(file)
    

def load(file, mmap_mode=None):
    ## load value
    value = np.load(file,mmap_mode=mmap_mode)
    ## unpack value if npz file with one variable
    try:
        value.keys
    except AttributeError:
        pass
    else:
        keys = value.keys()
        if len(keys) == 1:
            value = value[keys[0]]
    ## return
    return value


def save_txt(file, values, format_string=None, make_read_only=False, overwrite=False, create_path_if_not_exists=True):
    values = np.asarray(values)

    if len(values.shape) == 0:
        values = values.reshape(1)

    ## chose format string if not passed
    if format_string is None:
        if values.dtype == np.int:
            format_string = '%d'
        else:
            format_string = '%.18e'

    if create_path_if_not_exists:
        (dir, filename) = os.path.split(file)
        os.makedirs(dir, exist_ok=True)
    if overwrite:
        util.io.fs.remove_file(file, force=True, not_exist_okay=True)
    np.savetxt(file, values, fmt=format_string)
    if make_read_only:
        util.io.fs.make_read_only(file)


def load_txt(file):
    ## load values from file
    values = np.loadtxt(file)

    ## cast to int if possible
    values_int = values.astype(np.int)
    if (values_int == values).all():
        values = values_int

    ## if only one value return pure value
    if values.size == 1:
        values = values[0]

    return values


def save_np_and_txt(file, array, compressed=None, make_read_only=False, overwrite=False, create_path_if_not_exists=True, format_string='%.6g'):
    (file_without_ext, file_ext) = os.path.splitext(file)
    (dir, filename_without_ext) = os.path.split(file_without_ext)
    os.makedirs(dir, exist_ok=True)

    file_np = add_file_ext(file_without_ext, compressed=compressed)
    file_txt = file_without_ext + '.txt'

    array = np.asarray(array)

    save(file_np, array, compressed=compressed, make_read_only=make_read_only, overwrite=overwrite, create_path_if_not_exists=create_path_if_not_exists)
    save_txt(file_txt, array, format_string=format_string, make_read_only=make_read_only, overwrite=overwrite, create_path_if_not_exists=create_path_if_not_exists)



import os
import re
import subprocess

import numpy as np

import util.batch.general.system

import util.logging
logger = util.logging.logger



## batch setup

class BatchSystem(util.batch.general.system.BatchSystem):

    def __init__(self):
        from util.batch.rz.constants import COMMANDS, QUEUES, MAX_WALLTIME, MODEL_RENAMING, NODE_INFOS
        super().__init__(COMMANDS, QUEUES, max_walltime=MAX_WALLTIME, module_renaming=MODEL_RENAMING, node_infos=NODE_INFOS)


    def __str__(self):
        return 'RZ batch system'


    def _get_job_id_from_submit_output(self, submit_output):
        return submit_output


    def job_state(self, job_id):
        ## remove suffix from job id
        SUFFIX = '.rz.uni-kiel.de'
        if job_id.endswith(SUFFIX):
            job_id = job_id[:-len(SUFFIX)]

        ## get state of job
        process = subprocess.Popen((self.status_command, '-a', job_id), stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        qstat_result = process.communicate()[0].decode("utf-8")
        qstat_returncode = process.returncode

        logger.debug('qstat result: {} with exit code: {}'.format(qstat_result, qstat_returncode))

        ## 255 => cannot connect to server
        if qstat_returncode == 255:
            raise ConnectionError(qstat_result)

        return qstat_returncode


    def is_job_running(self, job_id):
        qstat_returncode = self.job_state(job_id)
        return qstat_returncode == 0


    def is_job_finished(self, job_id):
        qstat_returncode = self.job_state(job_id)
        return qstat_returncode == 35 or qstat_returncode == 153


    ## node setups
    
    def _nodes_state_one_kind(self, kind):
        logger.debug('Getting nodes state for kind {}.'.format(kind))
    
        ## grep free nodes
        def grep_qnodes(expression):
            command = '{} | grep -E {}'.format(self.nodes_command, expression)
            try:
                grep_result = subprocess.check_output(command, shell=True).decode("utf-8")
            except subprocess.CalledProcessError as e:
                logger.warning('Command {} returns with exit code {} and output "{}"'.format(command, e.returncode, e.output.decode("utf-8")))
                grep_result = 'offline'
    
            return grep_result
    
        # 24 f_ocean Barcelona nodes (8 CPUs per node, 2.1 GHz) (f_ocean queue)
        if kind == 'f_ocean' or  kind == 'barcelona':
            grep_result = grep_qnodes('"rzcl05[1-9]|rzcl06[0-9]|rzcl07[0-4]"')
        # 12 f_ocean2 nodes (16 CPUs per node, 2.6 GHz) (f_ocean2 queue)
        elif kind == 'f_ocean2':
            grep_result = grep_qnodes('"rzcl26[2-9]|rzcl27[0-3]"')
        # 2 fobigmem nodes (32 CPUs per node, 2.6 GHz) (fobigmem queue)
        elif kind == 'fobigmem':
            grep_result = grep_qnodes('"rzcl28[7-8]"')
        # 18 Westmere-nodes (12 CPUs per node, 2.67 GHz)
        elif kind == 'westmere':
            grep_result = grep_qnodes('"rzcl17[8-9]|rzcl18[0-9]|rzcl19[0-5]"')
        # 26 AMD-Shanghai nodes (8 CPUs per node, 2.4 GHz)
        elif kind == 'shanghai':
            grep_result = grep_qnodes('"rzcl11[8-9]|rzcl1[2-3][0-9]|rzcl14[0-3]"')
        # 1 AMD-Shanghai nodes (16 CPUs per node, 2.4 GHz)
        elif kind == 'amd128':
            grep_result = grep_qnodes('"rzcl116"')
        # 1 AMD-Magny node (48 CPUs per node, 2.1 GHz)
        elif kind == 'amd256':
            grep_result = grep_qnodes('"rzcl200"')
        # Shanghai Ethernet nodes (8 CPUs per node, 2.4 GHz) (bio_ocean queue)
        elif kind == 'bio_ocean' or kind == 'shanghai-ethernet':
            grep_result = grep_qnodes('"rzcl07[5-9]|rzcl0[8-9][0-9]|rzcl10[0-9]|rzcl11[0-4]"')
        # Shanghai Infiniband nodes (8 CPUs per node, 2.4 GHz) (math queue)
        elif kind == 'math' or kind == 'shanghai-infiniband':
            grep_result = grep_qnodes('"rzcl11[8-9]|rzcl1[2-3][0-9]|rzcl14[0-3]"')
        else:
            raise ValueError('Unknown CPU kind: ' + kind)
    
        logger.debug(grep_result)
    
        ## extract free cpus and memory from grep result
        grep_result_lines = grep_result.splitlines()
    
        number_of_nodes = len(grep_result_lines)
        free_cpus = np.empty(number_of_nodes, dtype=np.uint32)
        free_memory = np.empty(number_of_nodes, dtype=np.uint32)
    
        # format: "rzcl179 (7/12) (41943040kb/49449316kb) (free) (1285234.rzcluster/6)"
        for i in range(number_of_nodes):
            grep_result_line = grep_result_lines[i]
    
            # check if node down
            if 'down' in grep_result_line or 'state-unknown' in grep_result_line or 'offline' in grep_result_line:
                free_cpus[i] = 0
                free_memory[i] = 0
            # calculte free cpus and memory
            else:
                grep_result_line_split = grep_result_line.split()
    
                grep_cpus = [int(int_str) for int_str in re.findall('\d+', grep_result_line_split[1])]
                grep_memory = [int(int_str) for int_str in re.findall('\d+', grep_result_line_split[2])]
    
                free_cpus[i] = grep_cpus[1] - grep_cpus[0]
                free_memory[i] = int(np.floor((grep_memory[1] - grep_memory[0]) / (1024**2)))
    
        return (free_cpus, free_memory)


    def _nodes_state(self):
        state = {}
        for kind in self.node_infos.kinds():
            state[kind] = self._nodes_state_one_kind(kind)
        return util.batch.general.system.NodesState(state)
    


BATCH_SYSTEM = BatchSystem()


## job

class Job(util.batch.general.system.Job):

    def __init__(self, output_dir, force_load=False):
        super().__init__(BATCH_SYSTEM, output_dir, force_load=force_load, max_job_name_len=15)


    def init_job_file(self, job_name, nodes_setup, queue=None):
        ## set queue if missing
        cpu_kind = nodes_setup.node_kind

        if cpu_kind in ('f_ocean', 'f_ocean2'):
            if queue is not None and queue != cpu_kind:
                logger.warning('Queue {1} not supported for CPU kind {2}. CPU kind changed to {2}'.format(queue, cpu_kind))
            queue = cpu_kind
        else:
            if queue is None:
                walltime_hours = nodes_setup.walltime
                if walltime_hours is None:
                    queue = 'medium'
                elif walltime_hours <= self.batch_system.max_walltime['express']:
                    queue = 'express'
                elif walltime_hours <= self.batch_system.max_walltime['small']:
                    queue = 'small'
                elif walltime_hours <= self.batch_system.max_walltime['medium']:
                    queue = 'medium'
                elif walltime_hours <= self.batch_system.max_walltime['long']:
                    queue = 'long'
                elif walltime_hours <= self.batch_system.max_walltime['para_low']:
                    queue = 'para_low'
                else:
                    raise ValueError('Walltime hours > {} are not supported.'.format(self.batch_system.max_walltime['para_low']))

        ## set cpu kind
        if cpu_kind == 'f_ocean2':
            cpu_kind = None
        if cpu_kind == 'shanghai':
            cpu_kind = 'all'

        ## super
        super().init_job_file(job_name, nodes_setup, queue=queue, cpu_kind=cpu_kind)


    def _make_job_file_header(self, use_mpi):
        content = []
        ## shell
        content.append('#!/bin/bash')
        content.append('')
        ## name
        content.append('#PBS -N {}'.format(self.options['/job/name']))
        ## output file
        if self.output_file is not None:
            content.append('#PBS -j oe')
            content.append('#PBS -o {}'.format(self.output_file))
        ## queue
        content.append('#PBS -q {}'.format(self.options['/job/queue']))
        ## walltime
        if self.walltime_hours is not None:
            content.append('#PBS -l walltime={:02d}:00:00'.format(self.walltime_hours))
        ## select
        if self.cpu_kind is not None:
            cpu_kind_select = '{}=true:'.format(self.cpu_kind)
        else:
            cpu_kind_select = ''
        content.append('#PBS -l select={:d}:{}ncpus={:d}:mem={:d}gb'.format(self.options['/job/nodes'], cpu_kind_select, self.options['/job/cpus'], self.options['/job/memory_gb']))
        ## return
        content.append('')
        content.append('')
        return os.linesep.join(content)


    def _make_job_file_modules(self, modules):
        content = []
        if len(modules) > 0:
            ## init module system
            content.append('. /usr/share/Modules/init/bash')
            ## system modules
            for module in modules:
                content.append('module load {}'.format(module))
            content.append('module list')
            content.append('')
            content.append('')
        return os.linesep.join(content)


## imports from util.batch.general.system

from util.batch.general.system import NodeSetup, NodeSetupIncompleteError, JobError, JobExitCodeError



import util.observable.base

import util.logging
logger = util.logging.logger



class ObservableDict(util.observable.base.Observable, dict):
    
    def __init__(self, observers=None):
        dict.__init__(self)
        util.observable.base.Observable.__init__(self, observers=observers)
    
    
    def __setitem__(self, key, value):
        try:
            old_value = self.__getitem__(key)
        except KeyError:
            super().__setitem__(key, value)
            self.notify_added(key, value)
        else:
            if value != old_value:
                super().__setitem__(key, value)
                self.notify_changed(key, value, old_value)
    
    
    def __delitem__(self, key):
        value = self.__getitem__(key)
        super().__delitem__(key)
        self.notify_removed(key, value)
    
    
    def clear(self):
        old_dict = self.copy()
        super().clear()
        for key, value in old_dict.items():
            self.notify_removed(key, value)
    
    
    def update(self, update_dict):
        changed_keys_values = []
        new_keys_values = []
        for key, value in update_dict.items():
            try:
                old_value = self[key]:
            except KeyError:
                new_keys_values.append((key, value))
            else:
                if value != old_value:
                    changed_keys_values.append((key, value, old_value))
        
        super().update(update_dict)
        
        for (key, value, old_value) in changed_keys_values:
            self.notify_changed(key, value, old_value)        
        for (key, value) in new_keys_values:
            self.notify_added(key, value)
    
    
    def setdefault(self, key, value=None):
        value_added = key not in self
        value = super().setdefault(key, value)
        if value_added:
            self.notify_added(key, value)
        return value

    
    def pop(self, k, value=None):
        value_removed = key in self
        value = super().pop(self, k, value)
        if value_removed:
            self.notify_removed(key, value)
        return value

    
    def popitem(self):
        key, value = dict.popitem(self)
        self.notify_removed(key, value)
        return key, value 
    


class DependencyObserver:
    
    def __init__(self, independent_observable_dict, dependent_observable_dict, independent_key=None, dependent_key=None):
        self.independent_observable_dict = independent_observable_dict
        self.dependent_observable_dict = dependent_observable_dict
        self.independent_key = independent_key
        self.dependent_key = dependent_key
        
        independent_observable_dict.add_observer(self)
    
    
    def notify_dependency(self, key):
        if self.independent_key is None or key == self.independent_key:
            logger.debug('{}: dependency activated: independent dict {}, independent key {}, dependent dict {}, dependent key {}.'.format(self.__class__.__name__, self.independent_observable_dict, self.independent_key, self.dependent_observable_dict, self.dependent_key))
            if dependent_key is not None:
                del self.dependent_observable_dict[self.dependent_key]
            else:
                self.dependent_observable_dict.clear()
    
    
    def notify_changed(self, key, new_value, old_value):
        self.notify_dependency(key)
    
    def notify_added(self, key, new_value):
        self.notify_dependency(key)
    
    def notify_removed(self, key, old_value):
        self.notify_dependency(key)





import os
import subprocess

import numpy as np

import util.batch.general.system

import util.logging
logger = util.logging.logger



## batch setup

class BatchSystem(util.batch.general.system.BatchSystem):

    def __init__(self):
        from util.batch.nec.constants import COMMANDS, QUEUES, MAX_WALLTIME, MODEL_RENAMING, NODE_INFOS
        super().__init__(COMMANDS, QUEUES, max_walltime=MAX_WALLTIME, module_renaming=MODEL_RENAMING, node_infos=NODE_INFOS)


    def __str__(self):
        return 'NEC batch system'


    def _get_job_id_from_submit_output(self, submit_output):
        # Output form: "Request 130530.ace-ssiox submitted to queue: clmedium."
        submit_output_splitted = submit_output.split(' ')
        assert len(submit_output_splitted) == 6
        assert submit_output_splitted[5][:-1] in self.queues
        job_id = submit_output_splitted[1]
        return job_id


    def job_state(self, job_id):
        ## get state of job
        output = subprocess.check_output((self.status_command, job_id)).decode("utf-8")
        logger.debug('qstat result:\n{}'.format(output.strip()))
        return output


    def is_job_running(self, job_id):
        output = self.job_state(job_id)
        return 'RequestID' in output


    def _nodes_state(self):
        output = subprocess.check_output(self.nodes_command).decode('utf8')
        lines = output.splitlines()
        
        state = {}
        
        for line in lines:
            line = line.strip()
            for node_kind in self.node_infos.kinds():
                if line.startswith(node_kind):
                    line_splitted = line.split(' ')
                    number_of_free_nodes = int(line_splitted[-1])
                    
                    if number_of_free_nodes < 0:
                        logger.warn('Number of free nodes in the following line is negative, setting free nodes to zero.\n{}'.format(line))
                        number_of_free_nodes = 0
                    
                    logger.debug('Extracting nodes states from line "{}": node kind {} with {} free nodes.'.format(line, node_kind, number_of_free_nodes))
                    free_cpus = np.ones(number_of_free_nodes, dtype=np.uint32) * self.node_infos.cpus(node_kind)
                    free_memory = np.ones(number_of_free_nodes, dtype=np.uint32) * self.node_infos.memory(node_kind)
                    
                    state[node_kind] = (free_cpus, free_memory)
        
        return util.batch.general.system.NodesState(state)

        

BATCH_SYSTEM = BatchSystem()


## job

class Job(util.batch.general.system.Job):

    def __init__(self, output_dir, force_load=False):
        from util.batch.nec.constants import EXCEEDED_WALLTIME_ERROR_MESSAGE
        super().__init__(BATCH_SYSTEM, output_dir, force_load=force_load, exceeded_walltime_error_message=EXCEEDED_WALLTIME_ERROR_MESSAGE)


    def init_job_file(self, job_name, nodes_setup, queue=None):
        ## set queue if missing
        if queue is not None and queue != nodes_setup.node_kind:
            logger.warn('Queue {} and cpu kind {} have to be the same. Setting Queue to cpu kind.'.format(queue, nodes_setup.node_kind))
        queue = nodes_setup.node_kind

        ## super
        super().init_job_file(job_name, nodes_setup, queue=queue, cpu_kind=None)


    def _make_job_file_header(self, use_mpi):
        content = []
        ## shell
        content.append('#!/bin/bash')
        content.append('')
        ## name
        content.append('#PBS -N {}'.format(self.options['/job/name']))
        ## output file
        if self.output_file is not None:
            content.append('#PBS -j o')
            content.append('#PBS -o {}'.format(self.output_file))
        ## queue
        content.append('#PBS -q {}'.format(self.options['/job/queue']))
        ## walltime
        if self.walltime_hours is not None:
            content.append('#PBS -l elapstim_req={:02d}:00:00'.format(self.walltime_hours))
        ## select
        content.append('#PBS -b {:d}'.format(self.options['/job/nodes']))
        content.append('#PBS -l cpunum_job={:d}'.format(self.options['/job/cpus']))
        content.append('#PBS -l memsz_job={:d}gb'.format(self.options['/job/memory_gb']))
        ## MPI
        if use_mpi:
            content.append('#PBS -T intmpi')
        ## return
        content.append('')
        content.append('')
        return os.linesep.join(content)


    def _make_job_file_modules(self, modules):
        content = []
        if len(modules) > 0:
            ## system modules
            for module in modules:
                content.append('module load {}'.format(module))
            content.append('module list')
            content.append('')
            content.append('')
        return os.linesep.join(content)


## imports from util.batch.general.system

from util.batch.general.system import NodeSetup, NodeSetupIncompleteError, JobError, JobExitCodeError



import numpy as np
import ctypes

import multiprocessing
import multiprocessing.pool

import util.parallel.universal

import util.logging
logger = util.logging.logger



## sharred array

def shared_array_generic(size_or_initializer, shape, dtype=np.float64):
    logger.debug('Creating shared array with shape {} and dtype {}.'.format(shape, dtype))

    ## convert numpy type to C type
    if dtype.type is np.int64:
        ctype = ctypes.c_int64
    elif dtype.type is np.int32:
        ctype = ctypes.c_int32
    elif dtype.type is np.int16:
        ctype = ctypes.c_int16
    elif dtype.type is np.float64:
        ctype = ctypes.c_double
    elif dtype.type is np.float32:
        ctype = ctypes.c_float
    elif dtype.type is np.float128:
        ctype = ctypes.c_longdouble
    elif dtype.type is np.bool_:
        ctype = ctypes.c_bool
    else:
        raise ValueError('Data type {} of array is not supported.'.format(dtype.type))
    logger.debug('Using ctype {}'.format(ctype))

    ## make shared array
    shared_array_base = multiprocessing.Array(ctype, size_or_initializer, lock=False)
    shared_array = np.frombuffer(shared_array_base, dtype)
    shared_array.shape = shape   # prevent copy

    ## return
    logger.debug('Shared array created.')
    return shared_array


def shared_array(array):
    if array is not None:
        logger.debug('Creating shared array from array.')
    #     shared_array = shared_array_generic(array.flat, array.shape, array.dtype, lock=lock)
        shared_array = shared_array_generic(array.size, array.shape, array.dtype) ## do not use initializer ->  it needs additional memory
        shared_array[:] = array[:]
    #     np.testing.assert_array_equal(shared_array, array)
        return shared_array
    else:
        return None


def shared_zeros(shape, dtype=np.float64):
    logger.debug('Creating shared zeros array.')
    return shared_array_generic(np.array(shape).prod(), shape, dtype=dtype)


def share_all_arrays(args):
    ## make list
    if args is not None:
        args_list = list(args)
    else:
        args_list = []

    ## share arrays
    for i in range(len(args_list)):
        arg = args_list[i]

        if type(arg) in (np.ndarray, np.core.memmap):
            logger.debug('Sharing array at index {}.'.format(i))
            args_list[i] = shared_array(arg)

    ## return args
    args = type(args)(args_list)
    return args



## map functions

def map_parallel(function, values, number_of_processes=None, chunksize=1):
    assert callable(function)

    logger.debug('Creating multiprocessing pool with {} processes and chunksize {}.'.format(number_of_processes, chunksize))

    with multiprocessing.pool.Pool(processes=number_of_processes) as pool:
        results = pool.map(function, values, chunksize=chunksize)
        results = tuple(results)

    logger.debug('Parallel calculation with {} results completed.'.format(len(results)))

    return results


def starmap_parallel(function, values, number_of_processes=None, chunksize=1):
    assert callable(function)

    logger.debug('Creating multiprocessing pool with {} processes and chunksize {}.'.format(number_of_processes, chunksize))

    with multiprocessing.pool.Pool(processes=number_of_processes) as pool:
        results = pool.starmap(function, values, chunksize=chunksize)
        results = tuple(results)

    logger.debug('Parallel calculation with {} results completed.'.format(len(results)))

    return results



## shared arguments

def create_array_with_shared_kargs(shape, function, **kargs):
    logger.debug('Creating array with shape {} with multiprocessing and {} shared kargs.'.format(shape, len(kargs)))

    ## prepare indices
    indices = np.ndindex(*shape)

    ## execute in parallel
    with GlobalKargs(**kargs):
        results = map_parallel(eval_with_global_kargs, indices)
        array = np.array(tuple(results))

    ## create array
    logger.debug('Calculation completed. Got {} results.'.format(len(array)))
    array = array.reshape(shape)
    return array



def map_parallel_with_args(function, indices, *args, number_of_processes=None, chunksize=1, share_args=True):
    logger.debug('Mapping function with {} args of types {} and share {} to values with multiprocessing.'.format(len(args), tuple(map(type, args)), share_args))

    ## execute in parallel
    if share_args:
#         args = share_all_arrays(args, lock=False)
#         values = util.parallel.universal.args_generator_with_indices(indices, (function,))
        with GlobalArgs(function, *args):
            results = map_parallel(eval_with_global_args, indices)
    else:
        values = util.parallel.universal.args_generator_with_indices(indices, args)
        results = starmap_parallel(function, values)

    logger.debug('Parallel multiprocessing calculation with {} results completed.'.format(len(results)))

    return results




## global args

class MultipleUseError(Exception):

    def __init__(self, global_variable):
        self.global_variable = global_variable

    def __str__(self):
        return 'The global variable is already used for multiprocessing computation. Its content is {}'.format(self.global_variable)


class GlobalKargs:
    def __init__(self, f, **kargs):
        ## store kargs
        self.f = f
        self.kargs = kargs

        ## init global variable
        global _global_kargs
        try:
            _global_kargs
        except NameError:
            _global_kargs = None

    def __enter__(self):
        logger.debug('Storing {} global kargs of types {}.'.format(len(self.kargs), tuple(map(type, self.kargs))))

        ## store global variable
        global _global_kargs
        global _global_kargs_f
        if _global_kargs is None:
            _global_kargs = self.kargs
            _global_kargs_f = self.f
        else:
            raise MultipleUseError(_global_kargs)

    def __exit__(self, exc_type, exc_value, traceback):
        logger.debug('Deleting global kargs.')

        ## del global variable
        if exc_type is not MultipleUseError:
            global _global_kargs
            global _global_kargs_f
            _global_kargs = None
            _global_kargs_f = None


class GlobalArgs:
    def __init__(self, f, *args):
        ## store args
        self.f = f
        self.args = args

        ## init global variable
        global _global_args
        try:
            _global_args
        except NameError:
            _global_args = None

    def __enter__(self):
        logger.debug('Storing {} global args of types {}.'.format(len(self.args), tuple(map(type, self.args))))

        ## store global vari_global_args_fable
        global _global_args
        global _global_args_f
        if _global_args is None:
            _global_args = self.args
            _global_args_f = self.f
        else:
            raise MultipleUseError(_global_args)

    def __exit__(self, exc_type, exc_value, traceback):
        logger.debug('Deleting global args.')

        ## del global variable
        if exc_type is not MultipleUseError:
            global _global_args
            global _global_args_f
            _global_args = None
            _global_args_f = None


def eval_with_global_kargs(i, f):
    global _global_kargs
    return f(i, **_global_kargs)


def eval_with_global_args(i):
    global _global_args
    global _global_args_f
    return _global_args_f(i, *_global_args)












##############################
# import numpy as np
# import ctypes
#
# import multiprocessing
# import multiprocessing.pool
#
# import util.parallel.universal
#
# import util.logging
# logger = util.logging.logger
#
#
#
# ## sharred array
#
# def shared_array_generic(size_or_initializer, shape, dtype=np.float64, lock=True):
#     logger.debug('Creating shared array with shape {}, dtype {} and lock {}.'.format(shape, dtype, lock))
#
#     ## convert numpy type to C type
#     if dtype.type is np.int64:
#         ctype = ctypes.c_int64
#     elif dtype.type is np.int32:
#         ctype = ctypes.c_int32
#     elif dtype.type is np.int16:
#         ctype = ctypes.c_int16
#     elif dtype.type is np.float64:
#         ctype = ctypes.c_double
#     elif dtype.type is np.float32:
#         ctype = ctypes.c_float
#     elif dtype.type is np.float128:
#         ctype = ctypes.c_longdouble
#     elif dtype.type is np.bool_:
#         ctype = ctypes.c_bool
#     else:
#         raise ValueError('Data type {} of array is not supported.'.format(dtype.type))
#     logger.debug('Using ctype {}'.format(ctype))
#
#     ## make shared array
#     shared_array_base = multiprocessing.Array(ctype, size_or_initializer, lock=lock)
#     shared_array = np.frombuffer(shared_array_base, dtype)
#     shared_array.shape = shape   # prevent copy
#
#     ## return
#     logger.debug('Shared array created.')
#     return shared_array
#
#
# def shared_array(array, lock=True):
#     if array is not None:
#         logger.debug('Creating shared array from array.')
#     #     shared_array = shared_array_generic(array.flat, array.shape, array.dtype, lock=lock)
#         shared_array = shared_array_generic(array.size, array.shape, array.dtype, lock=lock) ## do not use initializer ->  it needs additional memory
#         shared_array[:] = array[:]
#     #     np.testing.assert_array_equal(shared_array, array)
#         return shared_array
#     else:
#         return None
#
#
# def shared_zeros(shape, dtype=np.float64, lock=True):
#     logger.debug('Creating shared zeros array.')
#     return shared_array_generic(np.array(shape).prod(), shape, dtype=dtype, lock=lock)
#
#
# def share_all_arrays(args, lock=True):
#     ## make list
#     if args is not None:
#         args_list = list(args)
#     else:
#         args_list = []
#
#     ## share arrays
#     for i in range(len(args_list)):
#         arg = args_list[i]
#
#         if type(arg) in (np.ndarray, np.core.memmap):
#             logger.debug('Sharing array at index {}.'.format(i))
#             args_list[i] = shared_array(arg, lock=lock)
#
#     ## return args
#     args = type(args)(args_list)
#     return args
#
#
#
# ## map functions
#
# def map_parallel(function, values, number_of_processes=None, chunksize=1):
#     assert callable(function)
#
#     logger.debug('Creating multiprocessing pool with {} processes and chunksize {}.'.format(number_of_processes, chunksize))
#
#     with multiprocessing.pool.Pool(processes=number_of_processes) as pool:
#         results = pool.map(function, values, chunksize=chunksize)
#         results = tuple(results)
#
#     logger.debug('Parallel calculation with {} results completed.'.format(len(results)))
#
#     return results
#
#
# def starmap_parallel(function, values, number_of_processes=None, chunksize=1):
#     assert callable(function)
#
#     logger.debug('Creating multiprocessing pool with {} processes and chunksize {}.'.format(number_of_processes, chunksize))
#
#     with multiprocessing.pool.Pool(processes=number_of_processes) as pool:
#         results = pool.starmap(function, values, chunksize=chunksize)
#         results = tuple(results)
#
#     logger.debug('Parallel calculation with {} results completed.'.format(len(results)))
#
#     return results
#
#
# ## shared arguments
#
# class MultipleUseError(Exception):
#
#     def __init__(self, global_variable):
#         self.global_variable = global_variable
#
#     def __str__(self):
#         return 'The global variable is already used for multiprocessing computation. Its content is {}'.format(self.global_variable)
#
#
# class GlobalKargs:
#     def __init__(self, **kargs):
#         ## store kargs
#         self.kargs = kargs
#
#         ## init global variable
#         global _global_kargs
#         try:
#             _global_kargs
#         except NameError:
#             _global_kargs = None
#
#     def __enter__(self):
#         logger.debug('Storing {} global kargs of types {}.'.format(len(self.kargs), tuple(map(type, self.kargs))))
#
#         ## store global variable
#         global _global_kargs
#         if _global_kargs is None:
#             _global_kargs = self.kargs
#         else:
#             raise MultipleUseError(_global_kargs)
#
#     def __exit__(self, exc_type, exc_value, traceback):
#         logger.debug('Deleting global kargs.')
#
#         ## del global variable
#         if exc_type is not MultipleUseError:
#             global _global_kargs
#             _global_kargs = None
#
#
# class GlobalArgs:
#     def __init__(self, *args):
#         ## store args
#         self.args = args
#
#         ## init global variable
#         global _global_args
#         try:
#             _global_args
#         except NameError:
#             _global_args = None
#
#     def __enter__(self):
#         logger.debug('Storing {} global args of types {}.'.format(len(self.args), tuple(map(type, self.args))))
#
#         ## store global variable
#         global _global_args
#         if _global_args is None:
#             _global_args = self.args
#         else:
#             raise MultipleUseError(_global_args)
#
#     def __exit__(self, exc_type, exc_value, traceback):
#         logger.debug('Deleting global args.')
#
#         ## del global variable
#         if exc_type is not MultipleUseError:
#             global _global_args
#             _global_args = None
#
#
# def eval_with_global_kargs(i, f):
#     global _global_kargs
#     return f(i, **_global_kargs)
#
# # def eval_with_global_kargs_zipped(args):
# #     return eval_with_global_kargs(*args)
#
# def eval_with_global_args(i, f):
#     global _global_args
#     return f(i, *_global_args)
#
# # def eval_with_global_args_zipped(args):
# #     return eval_with_global_args(*args)
#
#
#
#
#
# def create_array_with_shared_kargs(shape, function, **kargs):
#     logger.debug('Creating array with shape {} with multiprocessing and {} shared kargs.'.format(shape, len(kargs)))
#
# #     ## prepare input args
# #     values = util.parallel.universal.args_generator_with_shape(shape, (eval_with_global_kargs, function))
# #
# #     ## execute in parallel
# #     with GlobalKargs(**kargs):
# #         results = map_parallel(util.parallel.universal.eval_function_with_index_and_args, values)
#
#     ## prepare input args
#     values = util.parallel.universal.args_generator_with_shape(shape, (function,))
#
#     ## execute in parallel
#     with GlobalKargs(**kargs):
#         results = starmap_parallel(eval_with_global_args, values)
#
#     ## create array
#     array = np.array(list(results))
#     logger.debug('Calculation completed. Got {} results.'.format(len(array)))
#     array = array.reshape(shape)
#     return array
#
#
# # def create_array_with_shared_args(shape, function, *args):
# #     logger.debug('Creating array with shape {} with multiprocessing and shared args.'.format(shape))
# #
# #     ## prepare input args
# #     values = args_generator(shape, (function,))
# #
# #     ## execute in parallel
# #     with GlobalArgs(*args):
# #         results = map_parallel(eval_with_global_args_zipped, values)
# #
# #     ## create array
# #     array = np.array(list(results))
# #     logger.debug('Calculation completed. Got {} results.'.format(len(array)))
# #     array = array.reshape(shape)
# #     return array
#
#
#
# # def map_parallel_with_args(function, values, *args, number_of_processes=None, chunksize=1, share_args=True):
# #     logger.debug('Mapping function with {} args and share {} to values.'.format(len(args), share_args))
# #
# #     ## execute in parallel
# #     if share_args:
# #         values = util.parallel.universal.args_generator_with_indices(values, (function))
# #         with GlobalArgs(*args):
# #             results = map_parallel(eval_with_global_args_zipped, values)
# #     else:
# #         values = util.parallel.universal.args_generator_with_indices(values, args)
# #         results = map_parallel(function, values)
# #
# #     return results
#
# # def map_parallel_with_args(function, values, *args, number_of_processes=None, chunksize=1, share_args=True):
# #     logger.debug('Mapping function with {} args of types {} and share {} to values.'.format(len(args), map(type, args), share_args))
# #
# #     ## execute in parallel
# #     if share_args:
# #         values = util.parallel.universal.args_generator_with_indices(values, (eval_with_global_args, function))
# #         with GlobalArgs(*args):
# #             results = map_parallel(util.parallel.universal.eval_function_with_index_and_args, values)
# #     else:
# #         values = util.parallel.universal.args_generator_with_indices(values, (function,) + args)
# #         results = map_parallel(util.parallel.universal.eval_function_with_index_and_args, values)
# #
# #     return results
#
# def map_parallel_with_args(function, indices, *args, number_of_processes=None, chunksize=1, share_args=True):
#     logger.debug('Mapping function with {} args of types {} and share {} to values with multiprocessing.'.format(len(args), tuple(map(type, args)), share_args))
#
#     ## execute in parallel
#     if share_args:
# #         args = share_all_arrays(args, lock=False)
#         values = util.parallel.universal.args_generator_with_indices(indices, (function,))
#         with GlobalArgs(*args):
#             results = starmap_parallel(eval_with_global_args, values)
#     else:
#         values = util.parallel.universal.args_generator_with_indices(indices, args)
#         results = starmap_parallel(function, values)
#
#     return results
#
#
#
# # def create_array_with_args(shape, function, *args, share_args=True):
# #     logger.debug('Creating array with shape {} with multiprocessing, {} args and share {}.'.format(shape, len(args), share_args))
# #
# #
# #     ## create array
# #     array = np.array(list(results))
# #     logger.debug('Calculation completed. Got {} results.'.format(len(array)))
# #     array = array.reshape(shape)
# #     return array
#
#
#
#
#
#
#
#
# # def create_array(shape, function, function_args=None, function_args_first=True, number_of_processes=None, chunksize=1):
# #     logger.debug('Creating array with shape {} in parallel with multiprocessing.'.format(shape))
# #
# #     ## create indices
# #     indices = np.ndindex(*shape)
# #     if function_args is None:
# #         function_args = ()
# #     if function_args_first:
# #         indices = [function_args + (i,) for i in indices]
# #     else:
# #         indices = [(i,) + function_args for i in indices]
# #
# #     ## execute in parallel
# #     results = starmap(function, indices, number_of_processes=number_of_processes, chunksize=chunksize)
# #
# #     ## create array
# #     array = np.array(results).reshape(shape)
# #     return array
#
#
#
# #
# #
# #
# # #import multiprocessing.pool
# # #import logging
# # #
# # # # def map(function, values, number_of_processes=None, chunksize=1):
# # # #     process_pool = multiprocessing.pool.Pool(processes=number_of_processes)
# # # #     result = process_pool.map(function, values, chunksize=chunksize)
# # # #     process_pool.close()
# # # #     process_pool.join()
# # # #     return result
# # #
# # def map(function, values, number_of_processes=None, chunksize=1):
# #     with multiprocessing.pool.Pool(processes=number_of_processes) as pool:
# #         result = pool.map(function, values, chunksize=chunksize)
# #     pool.join()
# #     return result
# #
# #
# # import numpy as np
# # import multiprocessing
# # import multiprocessing.pool
# # import ctypes
# #
# #
# # logger = multiprocessing.get_logger()
# #
# # def shared_zeros(shape, lock=True):
# #     ## create a shared numpy array
# #     shape = np.asarray(shape)
# #     size = int(shape.prod())
# #
# #     shared_array_base = multiprocessing.Array(ctypes.c_double, size, lock=lock)
# #     if lock:
# #         shared_array_base_obj = shared_array_base.get_obj()
# #     else:
# #         shared_array_base_obj = shared_array_base
# #     shared_array = np.frombuffer(shared_array_base_obj)
# #     shared_array = shared_array.reshape(shape)
# #
# #     logger.debug('Shared array with shape {} and lock {} created.'.format(shared_array, lock))
# #
# #     return shared_array
# #
# #
# #
# #
# # def create_array_in_parallel(shape, function, iterable, number_of_processes=None, chunksize=1):
# #     assert callable(function)
# #
# #     global shared_array
# #     shared_array = shared_zeros(shape, lock=False)
# #
# #     logger.debug('Creating pool with {} processes and chunksize.'.format(number_of_processes, chunksize))
# #
# #     pool = multiprocessing.pool.Pool(processes=number_of_processes)
# #     pool.map_async(function, iterable, chunksize=chunksize)
# #     pool.close()
# #     pool.join()
# #
# #     logger.debug('Parallel calculation completed.')
# #
# #     array = np.array(shared_array, copy=True)
# #     return array
# #
# #
# #
# # def compute(index):
# #     logger.debug('Calculating for index {}.'.format(index))
# #     shared_array[:,:, index] = index
# #
# #
# # def new_main():
# #     shape = (2,3,4)
# #     iterable = range(4)
# #     number_of_processes = 2
# #     chunksize = 2
# #     array = create_array_in_parallel(shape, compute, iterable, number_of_processes=number_of_processes, chunksize=chunksize)
# #     return array
# #
# #
# # if __name__ == '__main__':
# #     print(new_main())
# #
# # # def compute_2(array, index):
# # #     print('Setting index {} at array with shape {}'.format(index, array.shape))
# # #     array[:,:, index] = index
# # #
# # # def new_main_2():
# # #     shape = (2,3,4)
# # #     iterable = range(4)
# # #     number_of_processes = 2
# # #     chunksize = 2
# # #     function = compute_2
# # #
# # #     shared_array = shared_zeros(shape, lock=False)
# # #
# # #     pool = multiprocessing.pool.Pool(processes=number_of_processes)
# # #     for i in iterable:
# # #         pool.apply_async(function, args=(shared_array, i))
# # #     pool.close()
# # #     pool.join()
# # #
# # #     array = np.array(shared_array, copy=True)
# # #     shared_array = None
# # #     return array
# #
# #
# #
# #
# #
# #
# # if __name__=='__main__':
# #     arr = new_main()
# #     print(arr)
# # #     arr = new_main_2()
# # #     print(arr)


import abc
import fcntl
import errno
import os
import stat
import time

import util.io.fs
import util.io.filelock.general

import util.logging
logger = util.logging.logger



class FileLock:

    def __init__(self, file, exclusive=True, timeout=None, sleep=0.5):
        self._lock_count = 0
        self._file = file
        self._want_exclusive = exclusive
        self._is_exclusive = exclusive
        self._timeout = timeout
        self._sleep = sleep
        self._fd = None
        logger.debug('{}: Initiating file lock with timeout {} and sleep {}.'.format(self, timeout, sleep))
    
    
    def __str__(self):
        if self._is_exclusive:
            return 'File lock (exclusive) for {}'.format(self.file)
        else:
            return 'File lock (shared) for {}'.format(self.file)


    @property
    def file(self):
        return self._file

    @property
    def lockfile(self):
        return self._file + '.lock'


    @property
    def exclusive(self):
        return self._is_exclusive
    
    @exclusive.setter
    def exclusive(self, want_exclusive):
        if self._want_exclusive != want_exclusive:
            if want_exclusive and self._lock_count > 0 and not self._is_exclusive:
                raise ValueError('Lock is acquired as shared lock. To change to exclusive lock, first release this lock.')
            self._want_exclusive = want_exclusive
            logger.debug('{}: exclusive changed to {}.'.format(self, self._want_exclusive))
    
    
    def is_locked(self, exclusive_is_okay=True, shared_is_okay=True):
        return self._lock_count > 0 and ((self._is_exclusive and exclusive_is_okay) or (not self._is_exclusive and shared_is_okay))
    
    def lock_object(self, exclusive=True):
        self.want_exclusive = exclusive
        return self


    ## acquire and release
    
    def _open_lockfile(self):
        if self._fd is None:
            ## prepare flags and mode
            open_flags = os.O_RDONLY | os.O_CREAT | os.O_CLOEXEC
            open_mode = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH | stat.S_IWOTH
            ## open
            try:
                self._fd = os.open(self.lockfile, open_flags, mode=open_mode)
            except OSError as e:
                if e.errno == errno.ESTALE:
                    logger.warning('{}: {}. Retrying to open lock file'.format(self, e))
                    self._open_lockfile()
                else:
                    raise
            else:
                logger.debug('{}: Lock file {} opened.'.format(self, self.lockfile))
    
    
    def _close_lockfile(self):
        ## close
        if self._fd is not None:
            os.close(self._fd)
            self._fd = None
            logger.debug('{}: Lock file {} closed.'.format(self, self.lockfile))
    
    
    def _lock_lockfile(self, exclusive=True, blocking=True):  
        assert self._fd is not None   
        assert self._lock_count == 0
           
        ## prepare lock flags
        if exclusive:
            lock_flags = fcntl.LOCK_EX
        else:
            lock_flags = fcntl.LOCK_SH
        if not blocking:
            lock_flags = lock_flags | fcntl.LOCK_NB
        ## lock
        fcntl.flock(self._fd, lock_flags)
        self._is_exclusive = exclusive
        self._lock_count = 1
        logger.debug('{}: Lock file {} locked with exclusive={}.'.format(self, self.lockfile, exclusive))
    
    
    def _unlock_lockfile(self):
        assert self._fd is not None
        ## unlock
        if self._lock_count > 0:
            fcntl.flock(self._fd, fcntl.LOCK_UN)
            self._lock_count = 0
            logger.debug('{}: Lock file {} unlocked.'.format(self, self.lockfile))
    

    def _acquire(self, exclusive=True, timeout=None):
        ## save start time for timeout
        if timeout is not None:
            start_time = time.time()
        
        ## save lock type
        self._is_exclusive = exclusive
        logger.debug('{}: Acquiring with timeout {}.'.format(self, timeout))
        
        has_lock = False
        try:
            while not has_lock:        
                ## open file
                self._open_lockfile()
                
                ## try to get lock
                try:
                    self._lock_lockfile(exclusive=exclusive, blocking=timeout is None)
                except BlockingIOError as e:
                    ## check if regular timeout
                    if e.errno != errno.EAGAIN or timeout is None:
                        logger.warning('{}: Retrying to get lock because an BlockingIOError occured: {}'.format(self, e))
                else:
                    ## if file was removed in between, open new file
                    if not util.io.fs.fd_is_file(self._fd, self.lockfile, not_exist_okay=True):
                        logger.debug('{}: Lock file {} was removed in beetween. Opening new lock file.'.format(self, self.lockfile))
                        self._unlock_lockfile()
                        self._close_lockfile()
                    ## lock successfull
                    else:
                        logger.debug('{}: Fresh acquired.'.format(self))
                        has_lock = True
                
                ## handle timout
                if not has_lock:
                    ## if timeout reached, raise FileLockTimeoutError
                    if timeout is not None and time.time() > (start_time + timeout):
                        logger.debug('{}: Could not be acquired. Timeout {} reached.'.format(self, timeout))
                        raise util.io.filelock.general.FileLockTimeoutError(self.lockfile, timeout) from e
                    ## else wait
                    else:
                        time.sleep(self._sleep)
        except:
            self._unlock_lockfile()
            self._close_lockfile()
            raise
        
        assert self._lock_count > 0
        assert self._fd is not None
        assert self.is_locked(exclusive_is_okay=exclusive, shared_is_okay=not exclusive)
    
    
    def acquire(self):
        if self._lock_count == 0:
            self._acquire(exclusive=self._want_exclusive, timeout=self._timeout)
        else:
            self._lock_count = self._lock_count + 1
        logger.debug('{}: Now aquired {} times. (One time added).'.format(self, self._lock_count))


    def _release(self):
        try:
        
            ## try to get exclusive lock
            if not self.is_locked(exclusive_is_okay=True, shared_is_okay=False):
                self._unlock_lockfile()
                try:
                    self._acquire(exclusive=True, timeout=0)
                    assert self.is_locked(exclusive_is_okay=True, shared_is_okay=False)
                except util.io.filelock.general.FileLockTimeoutError:
                    logger.debug('{}: Could not remove lock file {}. It is locked by another process.'.format(self, self.lockfile))
            
            ## if exclusive lock, remove lock file
            if self.is_locked(exclusive_is_okay=True, shared_is_okay=False):
                assert self._fd is not None
                assert self.lockfile is not None
                assert util.io.fs.fd_is_file(self._fd, self.lockfile, not_exist_okay=False)
                try:
                    os.remove(self.lockfile)
                except OSError as e:
                    if e.errno == errno.EBUSY:
                        logger.debug('{}: Could not remove lock file {}. It is used by another process.'.format(self, self.lockfile))
                    else:
                        raise
                else:
                    logger.debug('{}: Lock file {} removed.'.format(self, self.lockfile))
        
        ## cleanup
        finally:
            self._unlock_lockfile()
            self._close_lockfile()
        
        assert self._lock_count == 0
        assert self._fd is None
        assert not self.is_locked()
        logger.debug('{}: Entirely released.'.format(self))


    def release(self):
        if self._lock_count == 1:
            self._release()
        elif self._lock_count > 1:
            self._lock_count = self._lock_count - 1        
            logger.debug('{}: Now aquired {} times. (One time removed).'.format(self, self._lock_count))
        else:
            logger.debug('{}: Must not be released since it is not aquired.'.format(self))
            

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()
    
    def __del__(self):
        if self._lock_count > 1:
            self._release()
        



class LockedFile(FileLock):

    def __init__(self, file, cache_beyond_lock=True, timeout=None, sleep=0.5):
        logger.debug('Locked file {}: Creating lock file with cache_beyond_lock {}.'.format(file, cache_beyond_lock))
        super().__init__(file, timeout=timeout, sleep=sleep)
        self.cache_beyond_lock = cache_beyond_lock
        self.file_value = None
        self.cached_file_modified_time = None
    
    
    def _release(self):
        if not self.cache_beyond_lock:
            self.file_value = None
        super()._release()
    
    
    ## modified time
    
    def modified_time(self):
        return os.stat(self.file).st_mtime_ns
    
    ## cache functions
    
    def _cache_set_value(self, value):
        self.file_value = value
        if self.cache_beyond_lock:
            self.cached_file_modified_time = self.modified_time()
    
    def _cache_is_valid(self):
        return self.file_value is not None and (not self.cache_beyond_lock or self.cached_file_modified_time == self.modified_time())
        
    
    ## load
    
    @abc.abstractmethod
    def _load(self, file):
        pass
        
    def load(self):
        if not self._cache_is_valid():
            logger.debug('Locked file {}: Loading value.'.format(self.file))
            with self.lock_object(exclusive=False):
                value = self._load(self.file)
                self._cache_set_value(value)
            logger.debug('Locked file {}: Value loaded.'.format(self.file))
        else:
            value = self.file_value
        
        assert value is not None
        return value  


    ## save
    @abc.abstractmethod
    def _save(self, file, value):
        pass
    
    def save(self, value):
        logger.debug('Locked file {}: Saving content.'.format(self.file))
        with self.lock_object(exclusive=True):
            self._save(self.file, value)
            self._cache_set_value(value)
        logger.debug('Locked file {}: Content saved.'.format(self.file))
    



import os.path

import util.io.env

WORK_DIR = util.io.env.load('WORK')
HOME_DIR = util.io.env.load('HOME')
try:
    TMP_DIR = util.io.env.load('TMP')
except util.io.env.EnvironmentLookupError:
    TMP_DIR = os.path.join(WORK_DIR, 'tmp')


import util.index_database.fs_based
import util.petsc.universal


class Database(util.index_database.fs_based.Database):
    
    def __init__(self, value_dir, value_filenames, value_reliable_decimal_places=15, tolerance_options=None):
        ## call super constructor
        super().__init__(value_dir, value_filenames, value_reliable_decimal_places=value_reliable_decimal_places, tolerance_options=tolerance_options)
    
    
    ## load and save
    def _load_file(self, value_file):
        return util.petsc.universal.load_petsc_vec_to_numpy_array(value_file)
    
    def _save_file(self, value_file, value):
        util.petsc.universal.save_numpy_array_to_petsc_vec(value_file, value)




import numpy as np
from petsc4py import PETSc as petsc

import logging
logger = logging.getLogger(__name__)


## save petsc

def save_petsc(file, petsc_object):
    logger.debug('Saving petsc object to %s.', file)

    viewer = petsc.Viewer().createBinary(file, petsc.Viewer.Mode.WRITE)
    petsc_object.view(viewer)
    viewer.destroy()



## load petsc

def load_petsc_vec(file):
    logger.debug('Loading petsc vector from %s.', file)

    viewer = petsc.Viewer().createBinary(file, petsc.Viewer.Mode.READ)
    vec = petsc.Vec().load(viewer)
    viewer.destroy()

    return vec


def load_petsc_mat(file):
    logger.debug('Loading petsc matrix from %s.', file)

    viewer = petsc.Viewer().createBinary(file, petsc.Viewer.Mode.READ)
    mat = petsc.Mat().load(viewer)
    viewer.destroy()

    return mat



def petsc_vec_to_array(vec):
    logger.debug('Converting petsc vector to array.')
    array = np.array(vec, copy=True)

#     ownership_range = vec.getOwnershipRange()
#     size_local = ownership_range[1] - ownership_range[0]
#     size_global = vec.getSize()
#
#     array = np.empty(size_global, dtype=dtype)
#
#     for i_local in range(size_local):
#         i_global = y_ownership_range[0] + i_local


    return array


def petsc_mat_to_array(mat, dtype=float):
    logger.debug('Converting petsc matrix to array.')
    shape = mat.getSize()
    array = np.zeros(shape, dtype=dtype)

    #get values row by row since petsc matrix are sparce
    for i in range(shape[0]):
        row = mat.getRow(i)
        indices = row[0]
        values = row[1]
        array[i, indices] = values;

    return array



def load_petsc_vec_to_numpy_array(file):
    vec = load_petsc_vec(file)
    array = petsc_vec_to_array(vec, dtype=dtype)
    vec.destroy()

    return array


def load_petsc_mat_to_array(file, dtype=float):
    mat = load_petsc_mat(file)
    array = petsc_mat_to_array(mat, dtype=dtype)
    mat.destroy()

    return array



## create functions

def create_petsc_vec_by_function(function, n, finish_assembly=True):
    ## create vec
    logger.debug('Creating petsc vec of length %d from function.', n)

    vec = petsc.Vec()
    vec.createMPI(n, comm=petsc.COMM_WORLD)

    for i in range(n):
        vec.setValue(i, function(i))

    ## assembly
    logger.debug('Beginning assembly.')
    vec.assemblyBegin()

    if finish_assembly:
        logger.debug('Ending assembly.')
        vec.assemblyEnd()

    return vec


def create_petsc_mat_by_function(function, size, finish_assembly=True):
    ## convert size
    try:
        n = m = int(size)
    except TypeError:
        if len(size) == 1:
            n = m = size[0]
        elif len(size) == 2:
            n = size[0]
            m = size[1]
        else:
            raise TypeError('Size has to be an int or a tuple of length one or two, but its zype is %s.' % type(size))

    ## create mat
    logger.debug('Converting array to petsc mat.')

    mat = petsc.Mat()
    mat.createDense([n,m], comm=petsc.COMM_WORLD)

    for i, j in np.ndindex(n, m):
        mat.setValue(i, j, function(i,j))

    ## assembly
    logger.debug('Beginning assembly.')
    mat.assemblyBegin(petsc.Mat.AssemblyType.FINAL)

    if finish_assembly:
        logger.debug('Ending assembly.')
        mat.assemblyEnd(petsc.Mat.AssemblyType.FINAL)

    return mat



## array to petsc

def array_to_petsc_vec(array, finish_assembly=True):
    ## check array
    if len(array.shape) == 1:
        n = array.shape[0]
    else:
        raise ValueError('array has to be a vector, but it´s shape is %s.' % array.shape)

    ## create vec
    logger.debug('Converting array to petsc vec.')

    function = lambda i: array[i]
    vec = create_petsc_vec_by_function(function, n, finish_assembly=finish_assembly)

    return vec


def array_to_petsc_mat(array, finish_assembly=True):
    ## check array
    if len(array.shape) == 2:
        (n, m) = array.shape
    else:
        raise ValueError('array has to be a matrix, but it´s shape is %s.' % array.shape)

    ## create mat
    logger.debug('Converting array to petsc mat.')

    function = lambda i,j: array[i,j]
    mat = create_petsc_mat_by_function(function, (n, m), finish_assembly=finish_assembly)

    return mat



## other

def print_petsc(petsc_object):
    petsc_object.view(petsc.Viewer.STDOUT())




def solve_linear_equations(A, b, solver_type=petsc.KSP.Type.CG, monitor=True):
    n = b.getSize()

    logger.debug('Prepare for solving linear equation system of size %d with solver %s.' % (n, solver_type))

    if monitor:
        petsc_opts = petsc.Options()
        petsc_opts.setValue('ksp_monitor', True)

    x = petsc.Vec()
    x.createMPI(n, comm=petsc.COMM_WORLD)

    ksp = petsc.KSP()
    ksp.create(petsc.COMM_WORLD)
    ksp.setFromOptions()
    ksp.setOperators(A)
    ksp.setType(solver_type)

    pc = ksp.getPC()
    pc.setType(petsc.PC.Type.NONE)

    logger.debug('Solving linear equation system.')

    if monitor:
        ksp.view(petsc.Viewer.STDOUT())

    ksp.solve(b, x)

    x.assemblyBegin()
    x.assemblyEnd()

    pc.destroy()
    ksp.destroy()

    logger.debug('Linear equation system solved.')

    return x



class Matrix_Shell_Petsc:

    def __init__(self, entry_function):
        self.entry_function = entry_function

    @property
    def n(self):
        return self.covariance_model.n

    def mult(self, context, x, y):
        logger.debug('Multiplying petsc matrix with vector without explicit matrix.')

        ## copy x to local vec
        scatter, x_local = petsc.Scatter.toAll(x)
        scatter.scatterBegin(x, x_local)
        scatter.scatterEnd(x, x_local)
        scatter.destroy()


        ## set y values
        y_ownership_range = y.getOwnershipRange()
        y_size_local = y_ownership_range[1] - y_ownership_range[0]
        y_size_global = y.getSize()

        for i_local in range(y_size_local):
            i_global = y_ownership_range[0] + i_local

            ## compute value
            value = 0
            for j_global in range(y_size_global):
                value += self.entry_function(i_global, j_global) * x_local.getValue(j_global)

            y.setValue(i_global, value)
        y.assemblyBegin()
        y.assemblyEnd()

        ## destroy local copy
        x_local.destroy()


import argparse

import util.math.sparse.decompose.with_cholmod
import util.io.object

import util.logging

if __name__ == "__main__":
    ## configure arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('load_file', type=argparse.FileType('r'))
    parser.add_argument('save_files', type=argparse.FileType('w'), nargs='+')
    parser.add_argument('-o', '--ordering_method', default='default')
    parser.add_argument('-r', '--return_type', default='P_L')
    parser.add_argument('-d', '--debug', action='store_true', help='Print debug infos.')
    args = parser.parse_args()

    n = len(args.return_type.split('_'))
    if len(args.save_files) != n:
        raise ValueError('For return type {} are {} save files needed.'.format(return_type, n))

    with util.logging.Logger(disp_stdout=args.debug):
        A = util.io.object.load(args.load_file.name)
        decomposition = util.math.sparse.decompose.with_cholmod.cholesky(A, ordering_method=args.ordering_method, return_type=args.return_type)
        for i in range(n):
            util.io.object.save(args.save_files[i].name, decomposition[i])
